from . import priority_scheduling, schedule
from .schedule import (
    DataKey,
    CoordKey,
    Key,
    _K,
    Schedule,
    ExperimentId,
    SimultaneousNum,
    ExperimentDetail,
    ExperimentDetailMap,
    ExperimentIdStationIdPairsMap,
    default_station,
    empty,
    from_ndarrays,
    to_dataframe,
    filter_by_time_range,
    filter_by_time_ranges,
)

# TODO: adj module and/or func name so func 'priority_scheduling' can be exported here



#!/usr/bin/env python

"""#TODO

"""

import logging
import numpy as np

from .radar_controller import RadarController
from ..radar.scans import Beampark

logger = logging.getLogger(__name__)

class Static(RadarController):
    """Takes in a direction and creates a static radar in beam park mode."""

    META_FIELDS = RadarController.META_FIELDS + [
        "scan_type",
        "dwell",
    ]

    def __init__(
        self,
        radar,
        azimuth=0.0,
        elevation=90.0,
        r=np.linspace(300e3, 1000e3, num=10),
        meta=None,
        **kwargs
    ):
        super().__init__(radar.copy(), meta=meta, **kwargs)
        if self.meta["dwell"] is None:
            self.meta["dwell"] = 0.1

        self.scan = Beampark(azimuth=azimuth, elevation=elevation, dwell=self.meta["dwell"])
        self.r = r

        logger.info(f"Static:init")

        self.point_radar()
        self.turn_on(self.radar)

    def default_meta(self):
        dic = super().default_meta()
        dic["scan_type"] = self.scan.__class__
        return dic

    def point_radar(self):
        """Assumes t is not array"""

        t = 0.0
        point_rx_to_tx = []
        point_tx = []
        for tx in self.radar.tx:
            point = self.scan.ecef_pointing(t, tx)

            if len(point.shape) > 1:
                point_tx.append(point + tx.ecef[:, None])
                __ptx = point[:, :, None] * self.r[None, None, :] + tx.ecef[:, None, None]
                point_rx_to_tx.append(__ptx.reshape(3, __ptx.shape[1] * __ptx.shape[2]))
            else:
                point_tx.append(point + tx.ecef)
                point_rx_to_tx.append(point[:, None] * self.r[None, :] + tx.ecef[:, None])

            RadarController._point_station(tx, point_tx[-1])

        for rx in self.radar.rx:
            rx_point = []
            for txi, tx in enumerate(self.radar.tx):
                # < 200 meters apart = same location for pointing
                if np.linalg.norm(tx.ecef - rx.ecef) < 200.0:
                    __ptx = point_tx[txi]
                    if len(__ptx.shape) == 1:
                        __ptx = __ptx.reshape(3, 1)
                    rx_point.append(__ptx)
                else:
                    rx_point.append(point_rx_to_tx[txi])
            rx_point = np.concatenate(rx_point, axis=1)

            RadarController._point_station(rx, rx_point)

        RadarController.coh_integration(radar, self.meta["dwell"])

    def generator(self, t):
        for ti in range(len(t)):
            self.toggle_stations(t[ti], self.radar)
            yield self.radar, self.meta



import typing as t
from datetime import datetime, timedelta
import numpy as np
import numpy.typing as npt
from astropy.time import Time, TimeDelta
from sorts.types import (
    Datetime64_us,
    Float64_as_deg,
    Datetime_Like,
    Timedelta64_us,
    Timedelta_Like,
    NDArray_N,
)


def convert_to_relative_time(epoch: Time, times: Time | TimeDelta | NDArray_N) -> NDArray_N:
    """Convert absolute or relative times to relative time"""
    if isinstance(times, TimeDelta):
        tv = times.sec
    elif isinstance(times, Time):
        tv = (times - epoch).sec
    else:
        tv = times  # assume input is in seconds

    if not isinstance(tv, np.ndarray):
        tv = np.array([tv])
    return tv


def wrap_latitudes_longitudes(
    lat: npt.NDArray[Float64_as_deg], lon: npt.NDArray[Float64_as_deg]
) -> tuple[npt.NDArray[Float64_as_deg], npt.NDArray[Float64_as_deg]]:
    """
    Wrap latitudes and longitudes so that they stay within [-90, 90] and [-180, 180)

    Returns `(wrapped_latitudes, wrapped_longitudes)` tuple

    - When a latitude wraps, the corresponding longitude value is flipped (added 180deg)
    """

    lat_wrapped = ((lat + 90) % 180) - 90
    flips = (lat + 90) // 180
    lon_wrapped = (((lon + 180) + (flips * 180)) % 360) - 180

    return (lat_wrapped, lon_wrapped)


def wrap_azimuths_elevations(
    az: npt.NDArray[Float64_as_deg],
    el: npt.NDArray[Float64_as_deg],
    el_limit: t.Literal[90] | t.Literal[180] = 90,
    allow_negative_elevations=False,
) -> tuple[npt.NDArray[Float64_as_deg], npt.NDArray[Float64_as_deg]]:
    """
    Wrap azimuths and elevations so that they stay within [-180, 180) and [0, 90] (or [0, 180])

    Returns `(wrapped_azimuths, wrapped_elevations)` tuple

    - When an elevation wraps, the corresponding azimuth value is flipped (added 180deg)
    - `el_limit` controls whether elevations are wrapped to [0, 90] or [0, 180]
    - Negative elevations will trigger exception by default, controlled by `allow_negative_elevations`
    """

    # throw exception if there are negative elevation(s)
    if not allow_negative_elevations:
        neg_el_mask = el < 0
        first_neg_el_idx = np.argmax(neg_el_mask) if np.any(neg_el_mask) else None
        if first_neg_el_idx is not None:
            raise RuntimeError(
                f"There are negative elevation(s), which is invalid. e.g. at [{first_neg_el_idx}]: {el[first_neg_el_idx]}"
            )

    el = el % 180

    if el_limit == 180:
        return (az, el)

    else:
        el_wrapped = (el) % 90
        flips = (el) // 90
        az_wrapped = (((az + 180) + (flips * 180)) % 360) - 180

        return (az_wrapped, el_wrapped)


def to_datetime64_us(d: Datetime_Like) -> Datetime64_us:
    match d:
        case np.datetime64():
            return d.astype("datetime64[us]")
        case str() | datetime():
            return np.datetime64(d, "us")
        case Time():
            return t.cast(np.datetime64, d.to_value("datetime64")).astype("datetime64[us]")
        case _:
            raise RuntimeError(f"Convertion from {type(d)} to `datetime64[us]` is not supported.")


def to_pydatetime(d: Datetime_Like) -> datetime:
    match d:
        case datetime():
            return d
        case str():
            return np.datetime64(d, "us").astype(datetime)
        case np.datetime64():
            return d.astype(datetime)
        case Time():
            return t.cast(datetime, d.to_datetime())
        case _:
            raise RuntimeError(f"Convertion from {type(d)} to `datetime64[us]` is not supported.")


def to_timedelta64_us(d: Timedelta_Like) -> Timedelta64_us:
    match d:
        case np.timedelta64():
            return d.astype("timedelta64[us]")
        case int() | timedelta():
            return np.timedelta64(d, "us")
        case TimeDelta():
            return np.timedelta64(int(t.cast(np.float64, d.to_value("us"))), "us")
        case _:
            raise RuntimeError(f"Convertion from {type(d)} to `timedelta64[us]` is not supported.")


def assert_class_attributes_equal_to(cls, values: t.Sequence):
    """
    Class attributes on classes are used as container for string constant in this code base.

    This is helper function to assure the class attributes are equal to what we expect.
    """

    class_attributes = [
        v for k, v in vars(cls).items() if not k.startswith("__") and not callable(v)
    ]
    assert set(class_attributes) == set(values)

    return True



"""
Defines the NewType `Schedule` and functions for its functionalities
"""

from __future__ import annotations
import logging, typing as t
from dataclasses import dataclass
from functools import reduce
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from sorts import types, utils, radar


logger = logging.getLogger(__name__)

CoordKey = t.Literal[
    "multi_index", "start_time", "exp_num", "stn_num", "simult_num", "enu", "e", "n", "u"
]
DataKey = t.Literal["end_time", "pointing"]
Key = t.Literal[DataKey, CoordKey]


class _K:
    """Internal helper class for accessing string keys consistently"""

    multi_index: t.Final = "multi_index"
    start_time: t.Final = "start_time"
    exp_num: t.Final = "exp_num"
    stn_num: t.Final = "stn_num"
    simult_num: t.Final = "simult_num"
    enu: t.Final = "enu"
    e: t.Final = "e"
    n: t.Final = "n"
    u: t.Final = "u"
    end_time: t.Final = "end_time"
    pointing: t.Final = "pointing"


utils.assert_class_attributes_equal_to(_K, t.get_args(Key))

Schedule = t.NewType("Schedule", xr.Dataset)
"""
A xarray `Dataset` of:
  ```
  Dimensions:      (multi_index: n, enu: 3)
  Coordinates:
    * multi_index  (multi_index) object MultiIndex ('exp_num', 'stn_num', 'simult_num', 'start_time')
    * start_time   (multi_index) datetime64[us]
    * exp_num      (multi_index) int16
    * stn_num      (multi_index) int16
    * simult_num   (multi_index) int16
    * enu          (enu) 'e' 'n' 'u'
  Data variables:
      end_time     (multi_index) datetime64[us]
      pointing     (enu, multi_index) float64
  ```
"""

SimultaneousNum = int
"""An int16 that corresponds to the order in simultaneous pointings"""

ExperimentId = int
"""A unique int16 that identifies an experiment"""


@dataclass(kw_only=True)
class ExperimentDetail:
    id: ExperimentId

    coh_int_bandwidth: float
    ipp: float
    pulse_length: float
    power: float
    bandwidth: float
    duty_cycle: float
    noise_temp: float

    slice_duration: types.Timedelta64_us
    "Duration of a control slice, in micro-second"


ExperimentDetailMap = dict[ExperimentId, ExperimentDetail]


ExperimentIdStationIdPairsMap = dict[ExperimentId, list[tuple[radar.StationId, radar.StationId]]]


# TODO: this helper should ideally be part of radar module/subpackage,
#   but we can move it after that module is refactored
def default_station():
    return radar.Station(
        lat=0.0,
        lon=0.0,
        alt=0.0,
        min_elevation=0.0,
        beam=None,  # might not work when the `Station` typing is tightened
        uid=0,
    )


def empty() -> Schedule:
    multi_index = pd.MultiIndex.from_arrays(
        [
            np.empty(0, dtype=np.int16),
            np.empty(0, dtype=np.int16),
            np.empty(0, dtype=np.int16),
            np.empty(0, dtype="datetime64[us]"),
        ],
        names=(_K.exp_num, _K.stn_num, _K.simult_num, _K.start_time),
    )

    sch = xr.Dataset(
        coords={
            **xr.Coordinates.from_pandas_multiindex(multi_index, _K.multi_index),
            _K.enu: [_K.e, _K.n, _K.u],
        },
        data_vars={
            _K.end_time: (_K.multi_index, np.empty(0, dtype="datetime64[us]")),
            _K.pointing: ((_K.enu, _K.multi_index), np.empty((3, 0), dtype=np.float64)),
        },
    )

    return Schedule(sch)


def from_ndarrays(
    start_time: npt.NDArray[types.Datetime64_us],
    end_time: npt.NDArray[types.Datetime64_us],
    exp_num: npt.NDArray[np.int16],
    stn_num: npt.NDArray[np.int16],
    simult_num: npt.NDArray[np.int16],
    pointing: types.EnuCoordinates,
) -> Schedule:
    multi_index = pd.MultiIndex.from_arrays(
        [exp_num, stn_num, simult_num, start_time],
        names=(_K.exp_num, _K.stn_num, _K.simult_num, _K.start_time),
    )

    sch = xr.Dataset(
        coords={
            **xr.Coordinates.from_pandas_multiindex(multi_index, _K.multi_index),
            _K.enu: [_K.e, _K.n, _K.u],
        },
        data_vars={
            _K.end_time: (_K.multi_index, end_time),
            _K.pointing: ((_K.enu, _K.multi_index), pointing),
        },
    )

    return Schedule(sch)


# TODO: remove?
def to_dataframe(sch: Schedule) -> pd.DataFrame:
    df = pd.concat(
        t.cast(
            list[pd.DataFrame],
            [
                sch[_K.end_time].transpose().to_pandas(),
                sch[_K.pointing].transpose().to_pandas(),
            ],
        ),
        axis=1,
        copy=False,
    ).reset_index()

    return df


def filter_by_time_range(sch: Schedule, time_range: types.TimeRange_us) -> Schedule:
    mask = (sch[_K.start_time] >= time_range[0]) & (sch[_K.end_time] <= time_range[1])

    ds_masked = sch[{_K.multi_index: mask}]

    return ds_masked


def filter_by_time_ranges(sch: Schedule, time_ranges: t.Sequence[types.TimeRange_us]) -> Schedule:
    resultant_mask: xr.DataArray = reduce(
        xr.ufuncs.logical_or,
        [
            (sch[_K.start_time] >= time_range[0]) & (sch[_K.end_time] <= time_range[1])
            for time_range in time_ranges
        ],
    )

    ds_masked = sch[{_K.multi_index: resultant_mask}]

    return ds_masked



from __future__ import annotations
import logging, typing as t
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
import xarray as xr
import spacecoords
from sorts import radar, scheduling
from sorts.space_object import SpaceObject
from sorts.radar import Station
from sorts.types import (
    EcefStates,
    Datetime64_us,
    Timedelta64_us,
    Float64_as_sec,
    EnuCoordinates,
    Datetime_Like,
    Timedelta_Like,
)
from sorts.utils import to_datetime64_us, to_timedelta64_us
from .controller_base import ControllerBase

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class ControllerState:
    spobj_time: npt.NDArray[Datetime64_us]
    spobj_states: EcefStates

    @classmethod
    def empty(cls) -> t.Self:
        return cls(
            spobj_time=np.empty(0, dtype="datetime64[us]"),
            spobj_states=np.empty((6, 0), dtype=np.float64),
        )


class TrackerController(ControllerBase):
    """
    Generate pointing schedule that tracks a space object.

    - The preferred way to create instances of this class is via its class methods (e.g. `TrackerController.from_space_object`).
    - This class serve as a frontend to the `State` type in this module
    """

    ControllerState = ControllerState
    """shortcut to module attribute"""

    def __init__(
        self,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        exp_detail: scheduling.ExperimentDetail,
        spobj: SpaceObject | None,
        epoch: Datetime_Like | None,
        station_id_pairs: list[tuple[radar.StationId, radar.StationId]],
        state: ControllerState,
    ):
        """
        NOTE: This is intended as an internal constructor, please use the constructor methods to create instances.
        """

        self.tx_station = tx_station
        self.rx_stations = rx_stations
        self.exp_detail = exp_detail
        self.station_id_pairs = station_id_pairs
        self.spobj = spobj
        self.epoch = epoch

        self.state: ControllerState = state

    @classmethod
    def from_ecef_states(
        cls,
        time: npt.NDArray[Datetime64_us],
        space_object_states: EcefStates,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        exp_detail: scheduling.ExperimentDetail,
    ) -> t.Self:
        """A constructor method"""

        stn_pairs = [(tx_station.uid, rx_station.uid) for rx_station in rx_stations]

        ctrl = cls(
            tx_station=tx_station,
            rx_stations=rx_stations,
            exp_detail=exp_detail,
            spobj=None,
            epoch=None,
            station_id_pairs=stn_pairs,
            state=ControllerState(spobj_time=time, spobj_states=space_object_states),
        )

        return ctrl

    @classmethod
    def from_space_object(
        cls,
        spobj: SpaceObject,
        epoch: Datetime_Like,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        exp_detail: scheduling.ExperimentDetail,
    ) -> t.Self:
        """A constructor method"""

        stn_pairs = [(tx_station.uid, rx_station.uid) for rx_station in rx_stations]

        ctrl = cls(
            tx_station=tx_station,
            rx_stations=rx_stations,
            exp_detail=exp_detail,
            spobj=spobj,
            epoch=epoch,
            station_id_pairs=stn_pairs,
            state=ControllerState.empty(),
        )

        return ctrl

    def get_experiment_detail(self) -> scheduling.ExperimentDetail:
        return self.exp_detail

    def get_experiment_id_station_id_pairs_map(self) -> scheduling.ExperimentIdStationIdPairsMap:
        return {self.exp_detail.id: self.station_id_pairs}

    def get_station_map(self) -> dict[radar.StationId, radar.Station]:
        stn_map: dict[radar.StationId, radar.Station] = {}

        stn_map[self.tx_station.uid] = self.tx_station
        stn_map.update(list([(stn.uid, stn) for stn in self.rx_stations]))

        return stn_map

    def _compute_controller_state(
        self, start_time: Datetime_Like, end_time: Datetime_Like, slice_duration: Timedelta_Like
    ) -> ControllerState:
        """Do the computation and return the updated `state` property."""

        if self.spobj is None:
            raise RuntimeError(
                "Cannot compute space object ECEF states without `spobj` in the `spec` prop."
            )
        if self.epoch is None:
            raise RuntimeError(
                "Cannot compute space object ECEF states without `epoch` in the `spec` prop."
            )

        exp_detail: scheduling.ExperimentDetail = self.exp_detail

        # NOTE: for `np.arange` 'stop param,
        #   - we subtract 'slice_duration' so that only full slice are included
        #   - and add `+1` so that slice with time range `('end_time - 'slice_duration', 'end_time')` is included
        time: npt.NDArray[Datetime64_us] = np.arange(
            to_datetime64_us(start_time),
            to_datetime64_us(end_time) - to_timedelta64_us(slice_duration) + 1,
            exp_detail.slice_duration,
        )
        dt: npt.NDArray[Timedelta64_us] = time - to_datetime64_us(self.epoch)
        dsec = t.cast(npt.NDArray[Float64_as_sec], dt.astype(np.float64) / 1e6)

        ecefs = self.spobj.get_state(dsec)

        state = ControllerState(spobj_time=time, spobj_states=ecefs)

        return state

    def _generate(self) -> scheduling.Schedule:
        """Generate the schedules."""

        loc_zenith = np.array([0, 0, 1], dtype=np.float64)

        # generate pointings
        tx_pointings: EnuCoordinates = self.tx_station.enu(self.state.spobj_states[:3])

        tx_pointings_zenith_ang = spacecoords.linalg.vector_angle(
            loc_zenith, tx_pointings, degrees=True
        )
        tx_el_in_range_mask = tx_pointings_zenith_ang <= 90.0 - self.tx_station.min_elevation
        tx_pointings = tx_pointings[:, tx_el_in_range_mask]

        rxs_pointings: list[EnuCoordinates] = []
        rx_el_in_range_with_tx_masks: list[npt.NDArray[np.bool]] = []
        pure_rx_stations = [stn for stn in self.rx_stations if stn.uid != self.tx_station.uid]
        for rx_station in pure_rx_stations:
            rx_pointings: EnuCoordinates = rx_station.enu(self.state.spobj_states[:3])

            rx_pointings_zenith_ang = spacecoords.linalg.vector_angle(
                loc_zenith, rx_pointings, degrees=True
            )
            rx_el_in_range_mask = rx_pointings_zenith_ang <= 90.0 - rx_station.min_elevation

            rx_el_in_range_with_tx_mask = np.logical_and(tx_el_in_range_mask, rx_el_in_range_mask)
            rx_el_in_range_with_tx_masks.append(rx_el_in_range_with_tx_mask)

            rx_pointings = rx_pointings[:, rx_el_in_range_with_tx_mask]
            rxs_pointings.append(rx_pointings)

        tx_sch_time = self.state.spobj_time[tx_el_in_range_mask]
        tx_sch_len = len(tx_sch_time)

        tx_sch = scheduling.from_ndarrays(
            start_time=tx_sch_time,
            end_time=tx_sch_time + self.exp_detail.slice_duration,
            exp_num=np.full(tx_sch_len, self.exp_detail.id, dtype=np.int16),
            stn_num=np.full(tx_sch_len, self.tx_station.uid, dtype=np.int16),
            simult_num=np.full(tx_sch_len, 0, dtype=np.int16),
            pointing=tx_pointings,
        )

        rx_schs: list[scheduling.Schedule] = []
        for rx_stn, rx_mask, rx_pointings in zip(
            pure_rx_stations, rx_el_in_range_with_tx_masks, rxs_pointings
        ):
            rx_sch_time = self.state.spobj_time[rx_mask]
            rx_sch_len = len(rx_sch_time)

            rx_schs.append(
                scheduling.from_ndarrays(
                    start_time=rx_sch_time,
                    end_time=rx_sch_time + self.exp_detail.slice_duration,
                    exp_num=np.full(rx_sch_len, self.exp_detail.id, dtype=np.int16),
                    stn_num=np.full(rx_sch_len, rx_stn.uid, dtype=np.int16),
                    simult_num=np.full(rx_sch_len, 0, dtype=np.int16),
                    pointing=rx_pointings,
                )
            )

        resultant_sch = xr.concat([tx_sch, *rx_schs], dim=scheduling._K.multi_index)
        resultant_sch = resultant_sch.sortby(scheduling._K.start_time)
        output = resultant_sch

        return output

    def generate(
        self, start_time: Datetime_Like | None = None, end_time: Datetime_Like | None = None
    ) -> scheduling.Schedule:
        """
        Generate the schedules.
        `start_time` and `end_time` should be omitted if this instance is created from `TrackerController.from_ecef_states`
        """

        if start_time is not None and end_time is not None:
            self.state = self._compute_controller_state(
                start_time, end_time, self.exp_detail.slice_duration
            )

        sch = self._generate()

        return sch



#!/usr/bin/env python

"""Correlate measurement time series with a population of objects to find the best match.

Currently only works for Mono-static measurements.

# TODO: Assume a uniform prior distribution over population index, posterior distribution is the
probability of what object generated the data. Probability comes from measurement covariance.
"""

import logging
from tqdm import tqdm
import h5py
from astropy.time import Time

logger = logging.getLogger(__name__)

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
except ImportError:

    class COMM_WORLD:
        rank = 0
        size = 1

    comm = COMM_WORLD()

import numpy as np


def residual_distribution_metric(t, r, v, r_ref, v_ref, **kwargs):
    """Using the simulated and the measured ranges and rage-rates calculate a de-correlation metric.

    :param numpy.ndarray t: Times in seconds corresponding to measurement and object simulated data.
    :param numpy.ndarray r: Measured ranges in meters
    :param numpy.ndarray v: Measured rage-rates in meters per second
    :param numpy.ndarray r_ref: Object simulated ranges in meters
    :param numpy.ndarray v_ref: Object simulated rage-rates in meters per second
    :return: Metric value, smaller values indicate better match.
    :rtype: float
    """
    residual_r_mu = np.mean(r_ref - r)
    residual_v_mu = np.mean(v_ref - v)

    metric = np.abs(residual_r_mu) + np.abs(residual_v_mu)

    return metric


def generate_measurements(state_ecef, rx_ecef, tx_ecef):
    index_tuple = (slice(None),) + tuple(None for x in range(len(state_ecef.shape) - 1))

    r_tx = tx_ecef[index_tuple] - state_ecef[:3, :, ...]
    r_rx = rx_ecef[index_tuple] - state_ecef[:3, :, ...]

    r_tx_n = np.linalg.norm(r_tx, axis=0)
    r_rx_n = np.linalg.norm(r_rx, axis=0)
    r_sim = r_tx_n + r_rx_n

    v_tx = -np.sum(r_tx * state_ecef[3:, :, ...], axis=0) / r_tx_n
    v_rx = -np.sum(r_rx * state_ecef[3:, :, ...], axis=0) / r_rx_n

    v_sim = v_tx + v_rx

    return r_sim, v_sim


def within_fow(t, states, rx, tx):
    if len(states.shape) > 2:
        size = np.prod(states.shape[2:])
        shape = states.shape
        states.shape = states.shape[:2] + (size,)
        ok = [
            np.logical_and(
                rx.field_of_view(states[:, :, 0]),
                tx.field_of_view(states[:, :, 0]),
            )
        ]
        for j in range(1, size):
            ok += [
                np.logical_and(
                    rx.field_of_view(states[:, :, j]),
                    tx.field_of_view(states[:, :, j]),
                )
            ]
        ok = np.stack(ok, axis=1)
        states.shape = shape
        ok.shape = shape[1:]
    else:
        ok = np.logical_and(
            rx.field_of_view(states),
            tx.field_of_view(states),
        )
    return ok


def default_propagation_handling(obj, t, t_measurement_indices, measurements):
    return obj.get_state(t)


def save_measurements(hf, measurements, variables):
    meas = hf.create_group("measurements")
    meas.attrs["variables"] = variables
    for ind in range(len(measurements)):
        dat = meas.create_group(f"{ind}")
        for var in variables:
            dat[var] = measurements[ind][var]
        dat["t"] = measurements[ind]["t"]
        dat.attrs["epoch"] = measurements[ind]["epoch"].unix
        if measurements[ind]["tx"].uid is not None:
            dat.attrs["tx"] = measurements[ind]["tx"].uid
        if measurements[ind]["rx"].uid is not None:
            dat.attrs["rx"] = measurements[ind]["rx"].uid


def save_correlation_results(file, indecies, metric, measurements, variables, correlation_data):
    with h5py.File(file, "w") as hf:
        hf["indecies"] = indecies
        hf["metric"] = metric

        save_measurements(hf, measurements, variables)

        crl = hf.create_group("correlation_data")
        for key in correlation_data:
            grp = crl.create_group(f"{key}")
            for ind in range(len(correlation_data[key])):
                subgrp = grp.create_group(f"{ind}")
                subgrp.attrs["oid"] = correlation_data[key][ind]["oid"]
                subgrp.attrs["match"] = correlation_data[key][ind]["match"]
                subgrp["valid"] = correlation_data[key][ind]["valid"]
                for var in variables:
                    subgrp[f"{var}_ref"] = correlation_data[key][ind][f"{var}_ref"]


def load_measurements(hf):
    measurements = []
    meas = hf["measurements"]
    variables = meas.attrs["variables"]
    for key in meas:
        dat = {
            "t": meas[key]["t"][()],
            "epoch": Time(meas[key].attrs["epoch"], format="unix", scale="utc"),
        }
        if "tx" in meas[key].attrs:
            dat["tx"] = meas[key].attrs["tx"]
        if "rx" in meas[key].attrs:
            dat["rx"] = meas[key].attrs["rx"]
        for var in variables:
            dat[var] = meas[key][var][()]
        measurements.append(dat)
    return measurements, variables


def load_correlation_results(file):
    with h5py.File(file, "r") as hf:
        indecies = hf["indecies"][()]
        metric = hf["metric"][()]

        measurements, variables = load_measurements(hf)

        correlation_data = []
        crl = hf["correlation_data"]
        for key in crl:
            cdat = []
            grp = crl[key]
            for ind_str in grp:
                subgrp = grp[ind_str]
                dat = {
                    "oid": subgrp.attrs["oid"],
                    "match": subgrp.attrs["match"],
                    "valid": subgrp["valid"][()],
                }
                for var in variables:
                    dat[f"{var}_ref"] = subgrp[f"{var}_ref"][()]
                cdat.append(dat)
            correlation_data.append(cdat)

    return indecies, metric, measurements, correlation_data


def correlate(
    measurements,
    population,
    metric=residual_distribution_metric,
    metric_reduce=lambda x, y: x + y,
    forward_model=generate_measurements,
    sorting_function=lambda metric: np.argsort(metric, axis=0),
    valid_measurement_checker=within_fow,
    metric_dtype=np.float64,
    metric_fail_value=np.nan,
    propagation_handling=default_propagation_handling,
    variables=["r", "v"],
    meta_variables=[],
    n_closest=1,
    scalar_metric=True,
    MPI=False,
    save_states=False,
):
    """Given a mono-static measurement of ranges and rage-rates, a radar model and a population:
    correlate measurements with population.

    The behavior should be thought of as the following matrix

    Metric reduce |     Function     |       None
    Scalar metric
    -------------
                     Only one scalar   Each measurement
        True           correlation        correlated
                       calculated        individually
    -------------
                     Only one vector   Each data point
        False          correlation        correlated
                       calculated        individually
    -------------

    # TODO: Update docstring
    # TODO: Add FOV check option

    Parameters
    ----------
    measurements : list
        List of dictionaries that contains measurement data. Contents are described below.
    population : Population
        Population to correlate against.
    metric : function
        Metric used to correlate measurement and simulated object measurement.
    metric_reduce : function
        Metric used to correlate measurement and simulated object measurement.
        Can be `None`, in which case each measurement is correlated individually,
        for this to work the metric also needs to be vectorized.
    forward_model : function
        A pointer to a function that takes in the ecef-state, the rx and tx
        station ecefs and calculates the observed variables return as a tuple.
    sorting_function : function
        A pointer to a sorting function that takes in the metric result array
        and returns a list of indices indicating the sorting order.
    valid_measurement_checker : function
        A pointer to a function that checks if the measurnment to calculate a
        metric is valid or not. By default checks if the object is within both
        stations FOV. The function expects to take (time, states, rx station, tx station).
    propagation_handling : function
        A pointer to a function that handles the propagation of the object to
        the measurnment point. By defaults simple uses the space object
        `get_state` method. TODO: add doc reference here to the default function
        to use as template for modifications.
    metric_dtype : numpy.dtype
        A valid numpy dtype declaration for the metric output array.
        This allows complex structured results to be processed.
    variables : list
        The data variables recorded by the system. Theses should be in
        `measurements` and returned by the `forward_model`.
    meta_variables : list
        The data meta variables recorded by the system. These are input as
        keyword arguments to the metric and does not need to be produced by any model.
    n_closest : int
        Number of closest matches to save.
    scalar_metric : bool
        indicats if the metric returns a scalar or a vector. If `False` the
        `metric_reduce` is expected to only take one argument to reduce the vectorized results.
    MPI : bool
        If True use internal parallelization with MPI to calculate correlation.
        Turn to False to externally parallelize with MPI.
    save_states : bool
        If True use saves the propagated states in the output data.


    **Measurement data:**

      Each entry in the input :code:`measurements` list must be a dictionary
      that contains the following fields:
        * 't': [numpy.ndarray] Times relative epoch in seconds
        * 'epoch': [astropy.Time] epoch for measurements
        * 'tx': [sorts.TX] Pointer to the TX station
        * 'rx': [sorts.RX] Pointer to the RX station

      Then it will contain a entry for each name in the `variables` list. By default this is
        * 'r': [numpy.ndarray] Two-way ranges in meters
        * 'v': [numpy.ndarray] Two-way range-rates in meters per second

    """

    correlation_data = {}

    if n_closest > len(population):
        raise ValueError(
            f"Cannot generate n_closest={n_closest} greater than population size {len(population)}"
        )

    iters = []
    for rank in range(comm.size):
        iters.append(list(range(rank, len(population), comm.size)))
    pbar = tqdm(total=len(iters[comm.rank]), position=comm.rank)

    meas_num = 0
    for data in measurements:
        meas_num += len(data["t"])

    if scalar_metric:
        if metric_reduce is None:
            match_pop = np.empty((n_closest + 1, len(measurements)), dtype=metric_dtype)
            index_pop = np.empty((n_closest + 1, len(measurements)), dtype=np.int64)
        else:
            match_pop = np.empty((n_closest + 1,), dtype=metric_dtype)
            index_pop = np.empty((n_closest + 1,), dtype=np.int64)
    else:
        if metric_reduce is None:
            match_pop = np.empty((n_closest + 1, meas_num), dtype=metric_dtype)
            index_pop = np.empty((n_closest + 1, meas_num), dtype=np.int64)
        else:
            match_pop = np.empty((n_closest + 1,), dtype=metric_dtype)
            index_pop = np.empty((n_closest + 1,), dtype=np.int64)
            tmp_match_cache = np.empty((meas_num,), dtype=metric_dtype)

    # Just pick some reference epoch so we can shift time vectors
    # Then we shift relative the object later when we propagate
    ref_epoch = measurements[0]["epoch"]

    # Build unique time vector from all input measurements
    t_sorts = []
    t_selectors = []
    for di, data in enumerate(measurements):
        t = (data["epoch"] - ref_epoch).sec + data["t"]
        t_sort = t.argsort()
        t_sorts.append(t_sort)
        if di == 0:
            t_prop0 = t
            t_selectors = np.full(t.shape, di, dtype=np.int64)
            t_m_index = np.arange(len(t))
        else:
            t_prop0 = np.append(t_prop0, t)
            t_selectors = np.append(t_selectors, np.full(t.shape, di, dtype=np.int64))
            t_m_index = np.append(t_m_index, np.arange(len(t)))

    t_prop_base, t_prop_indices = np.unique(t_prop0, return_inverse=True)
    t_prop_args = np.argsort(t_prop_base)

    t_prop_args_i = np.empty_like(t_prop_args)
    t_prop_args_i[t_prop_args] = np.arange(t_prop_args.size)

    t_indices = []
    t_reverse_inds = np.arange(t_prop_args.size)
    t_reverse_inds = t_reverse_inds[t_prop_args_i][t_prop_indices]
    for tii in range(len(t_prop_base)):
        tii_matches = np.argwhere(tii == t_reverse_inds).flatten()
        t_indices.append([(t_selectors[x], t_m_index[x]) for x in tii_matches])

    iteration_counter = 0

    # Iterate trough population
    for ind in iters[comm.rank]:
        obj = population.get_object(ind)

        t_shift = (ref_epoch - obj.epoch).sec
        t_prop = t_prop_base + t_shift

        pbar.set_description("Correlating object {} ".format(ind))
        pbar.update(1)

        logger.debug("correlating {}/{}".format(ind + 1, len(population)))
        logger.debug(obj)

        object_correlation_data = []

        # get all states in one go
        # and mark which time corresponds to which measurnment set and which measurnment in that set
        states = propagation_handling(obj, t_prop[t_prop_args], t_indices, measurements)

        # correlate with forward model
        for di, data in enumerate(measurements):
            t = (data["epoch"] - obj.epoch).sec + data["t"][t_sorts[di]]
            tx = data["tx"]
            rx = data["rx"]
            # TODO: station data can be extraced and passed on more generally for
            # e.g. single station and multi station correlations
            tx_ecef = tx.ecef.copy()
            rx_ecef = rx.ecef.copy()

            # Select from all propagated states the ones for this
            # measurnment block in the same order as the measurnments
            states_data = states[:, t_prop_args_i, ...][:, t_prop_indices, ...][
                :, t_selectors == di, ...
            ]

            base_valid = valid_measurement_checker(t, states_data, rx, tx)
            if len(base_valid.shape) > 1:
                valid = np.sum(base_valid, axis=tuple(range(1, len(base_valid.shape)))) > 0
            else:
                valid = base_valid

            var = tuple(data[x][t_sorts[di]][valid] for x in variables)
            ref_var = forward_model(states_data[:, valid, ...], rx_ecef, tx_ecef)

            kwvar = {}
            for x in meta_variables:
                try:
                    kwvar[x] = data[x][t_sorts[di]][valid]
                except TypeError:
                    kwvar[x] = data[x]

            match = metric(t[valid], *(var + ref_var), **kwvar)

            if scalar_metric:
                if match.size == 0:
                    match = metric_fail_value
            else:
                _match = np.empty(t.shape, dtype=metric_dtype)
                _match[valid] = match
                _match[np.logical_not(valid)] = metric_fail_value
                match = _match

            # Calculate the metric, scalar or vectorized and using reduction if defined
            if scalar_metric:
                if metric_reduce is None:
                    match_pop[n_closest, di] = match
                else:
                    if di == 0:
                        match_pop[n_closest] = match
                    else:
                        match_pop[n_closest] = metric_reduce(
                            match_pop[n_closest],
                            match,
                        )
            else:
                if metric_reduce is None:
                    match_pop[n_closest, t_selectors == di] = match
                else:
                    tmp_match_cache[t_selectors == di] = match
                    if di == len(measurements) - 1:
                        match_pop[n_closest] = metric_reduce(tmp_match_cache)

            cdat = {}
            for i, x in enumerate(variables):
                if ref_var[i].size == 0:
                    ref_var[i].shape = (0,) + base_valid.shape[1:]
                _xp = np.full(base_valid.shape, np.nan, dtype=ref_var[i].dtype)
                _xp[valid, ...] = ref_var[i]
                cdat[f"{x}_ref"] = _xp

            cdat["match"] = match
            cdat["valid"] = valid
            if save_states:
                cdat["states"] = states_data
            cdat["oid"] = ind

            object_correlation_data.append(cdat)

        # Set population index for current metric values
        index_pop[n_closest, ...] = ind

        # Save current set of correlation data
        correlation_data[ind] = object_correlation_data

        # Get the current best matches for this rank
        if iteration_counter < n_closest:
            match_pop[iteration_counter, ...] = match_pop[n_closest, ...]
            index_pop[iteration_counter, ...] = index_pop[n_closest, ...]
        else:
            sorting_results = sorting_function(match_pop)
            match_pop = np.take_along_axis(match_pop, sorting_results, axis=0)
            index_pop = np.take_along_axis(index_pop, sorting_results, axis=0)

            # Walk trough saved correlation data and free unused
            kept_index_pop = index_pop[:n_closest, ...]
            for oid in list(correlation_data.keys()):
                if not np.any(oid == kept_index_pop):
                    del correlation_data[oid]

        iteration_counter += 1

    pbar.close()

    # Remove the last row used as cache
    match_pop = match_pop[:n_closest, ...]
    index_pop = index_pop[:n_closest, ...]

    # Communicate the best matches with root thread
    #  to gather results and select the best ones
    if MPI and comm.size > 1:
        matches_pop = [None] * comm.size
        indecies_pop = [None] * comm.size

        matches_pop[comm.rank] = match_pop
        indecies_pop[comm.rank] = index_pop

        if comm.rank == 0:
            logger.debug("Receiving all results <barrier>")

            for T in range(1, comm.size):
                matches_pop[T] = comm.recv(source=T, tag=T * 10 + 1)
                indecies_pop[T] = comm.recv(source=T, tag=T * 10 + 2)
                correlation_data.update(comm.recv(source=T, tag=T * 10 + 3))
                logger.debug(f"Received data from PID{T}")
        else:
            logger.debug("Distributing all results to process 0 <barrier>")

            comm.send(matches_pop[comm.rank], dest=0, tag=comm.rank * 10 + 1)
            comm.send(indecies_pop[comm.rank], dest=0, tag=comm.rank * 10 + 2)
            comm.send(correlation_data, dest=0, tag=comm.rank * 10 + 3)

        logger.debug("---> Distributing done </barrier>")

        if comm.rank == 0:
            match_pop = np.concatenate(matches_pop)
            index_pop = np.concatenate(indecies_pop)

            # Resort and filter after reciving the other MPI data
            sorting_results = sorting_function(match_pop)
            match_pop = np.take_along_axis(match_pop, sorting_results[:n_closest, ...], axis=0)
            index_pop = np.take_along_axis(index_pop, sorting_results[:n_closest, ...], axis=0)

            for oid in list(correlation_data.keys()):
                if not np.any(oid == index_pop):
                    del correlation_data[oid]
        else:
            correlation_data = None

        match_pop = comm.bcast(match_pop, root=0)
        index_pop = comm.bcast(index_pop, root=0)

    return index_pop, match_pop, correlation_data



#!/usr/bin/env python

"""Wrapper for the REBOUND propagator into SORTS format.
"""
import copy, pathlib, logging

import numpy as np
import scipy

import pyorb

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


try:
    import rebound
except ImportError:
    rebound = None

from astropy.time import Time

# Local import
from .base import Propagator
from .. import dates as dates
from .. import frames

logger = logging.getLogger(__name__)

class Rebound(Propagator):
    """Propagator class implementing the REBOUND propagator.

    Internally in the Rebound simulation a ICRS equivalent is always used but output is generated in HCRS.

    #TODO: add to this documentation
    """

    DEFAULT_MASSIVE = [
        "Sun",
        "Moon",
        "Mercury",
        "Venus",
        "Earth",
        "Mars",
        "Jupiter",
        "Saturn",
        "Uranus",
        "Neptune",
    ]
    DEFAULT_MASSES = [
        1.98855e30,
        7.34767309e22,
        0.330104e24,
        4.86732e24,
        5.97219e24,
        0.641693e24,
        1898.13e24,
        568.319e24,
        86.8103e24,
        102.410e24,
    ]

    DEFAULT_SETTINGS = copy.copy(Propagator.DEFAULT_SETTINGS)
    DEFAULT_SETTINGS.update(
        dict(
            out_frame="HCRS",
            in_frame="HCRS",
            integrator="IAS15",
            time_step=60.0,
            termination_check=False,
            termination_check_interval=1,
            massive_objects=DEFAULT_MASSIVE,
            save_massive_states=False,
            tqdm=False,
        )
    )

    def __init__(self, kernel, settings=None, **kwargs):
        self.sim = None
        assert rebound is not None, "Rebound python package not found"

        super(Rebound, self).__init__(settings=settings, **kwargs)
        logger.debug(f"sorts.propagator.Rebound:init")

        self.settings["massive_objects"] = [
            x.strip().capitalize() for x in self.settings["massive_objects"]
        ]

        self.kernel_path = kernel

        self.planets_mass = {
            key: val
            for key, val in zip(
                Rebound.DEFAULT_MASSIVE,
                Rebound.DEFAULT_MASSES,
            )
        }

        self.internal_frame = "HCRS"
        self.geo_internal_frame = "GCRS"
        self._earth_ind = self.planet_index("Earth")
        self._sun_ind = self.planet_index("Sun")
        self.N_massive = len(self.settings["massive_objects"])

    def __str__(self):
        from rebound import __version__, __build__

        s = ""
        s += "---------------------------------\n"
        s += "REBOUND version:     \t%s\n" % __version__
        s += "REBOUND built on:    \t%s\n" % __build__
        if self.sim is not None:
            s += "Number of particles: \t%d\n" % self.sim.N
            s += "Selected integrator: \t" + self.sim.integrator + "\n"
            s += "Simulation time:     \t%.16e\n" % self.sim.t
            s += "Current timestep:    \t%f\n" % self.sim.dt
        s += "---------------------------------"
        return s

    def planet_index(self, name):
        return [x.lower().strip() for x in self.settings["massive_objects"]].index(
            name.lower().strip()
        )

    def _setup_sim(self, epoch, init_massive_states=None):
        kpath = pathlib.Path(self.kernel_path)
        if init_massive_states is None:
            assert kpath.is_file(), f'Could not find "{kpath}" kernel file.'

        self.sim = rebound.Simulation()
        self.sim.units = ("m", "s", "kg")
        self.sim.integrator = self.settings["integrator"]

        if init_massive_states is None:
            bodies = self.settings["massive_objects"]
            if "Sun" not in bodies:
                bodies += ["Sun"]
            states = frames.get_solarsystem_body_states(
                bodies=bodies,
                epoch=epoch,
                kernel=str(kpath),
            )

            # Convert to HCRS
            for key in states:
                if key != "Sun":
                    states[key] -= states["Sun"]
            states["Sun"] -= states["Sun"]

        for i in range(len(self.settings["massive_objects"])):
            body = self.settings["massive_objects"][i]

            if init_massive_states is None:
                state = states[body]
            else:
                state = init_massive_states[:, i]

            self._add_state(state, self.planets_mass[body])

        self.sim.N_active = self.N_massive
        self.sim.dt = self.settings["time_step"]

    def _add_state(self, state, m):
        x, y, z, vx, vy, vz = state.flatten()
        self.sim.add(
            x=x,
            y=y,
            z=z,
            vx=vx,
            vy=vy,
            vz=vz,
            m=m,
        )

    def termination_check(self, t, step_index, massive_states, particle_states):
        raise NotImplementedError("Users need to implement this method to use termination checks")

    def _put_simulation_state(self, massive_states, particle_states, ti):
        if massive_states is not None:
            for p_n in range(self.N_massive):
                particle = self.sim.particles[p_n]
                massive_states[0, ti, p_n] = particle.x
                massive_states[1, ti, p_n] = particle.y
                massive_states[2, ti, p_n] = particle.z
                massive_states[3, ti, p_n] = particle.vx
                massive_states[4, ti, p_n] = particle.vy
                massive_states[5, ti, p_n] = particle.vz
        if particle_states is not None:
            for p_n in range(len(self.sim.particles) - self.N_massive):
                particle = self.sim.particles[self.N_massive + p_n]
                particle_states[0, ti, p_n] = particle.x
                particle_states[1, ti, p_n] = particle.y
                particle_states[2, ti, p_n] = particle.z
                particle_states[3, ti, p_n] = particle.vx
                particle_states[4, ti, p_n] = particle.vy
                particle_states[5, ti, p_n] = particle.vz
        return massive_states, particle_states

    def _get_helio_state(self):
        sun_state = np.zeros((6,), dtype=np.float64)
        sun = self.sim.particles[self._sun_ind]
        sun_state[0] = sun.x
        sun_state[1] = sun.y
        sun_state[2] = sun.z
        sun_state[3] = sun.vx
        sun_state[4] = sun.vy
        sun_state[5] = sun.vz
        return sun_state

    def _get_earth_state(self):
        earth_state = np.zeros((6,), dtype=np.float64)
        earth = self.sim.particles[self._earth_ind]
        earth_state[0] = earth.x
        earth_state[1] = earth.y
        earth_state[2] = earth.z
        earth_state[3] = earth.vx
        earth_state[4] = earth.vy
        earth_state[5] = earth.vz
        return earth_state

    def propagate(self, t, state0, epoch, **kwargs):
        """Propagate a state

        #TODO: add possible different equinox?
        #TODO: UPDATE THIS DOCSTRING
        """

        logger.debug(f"Rebound:propagate:len(t) = {len(t)}")

        t, epoch = self.convert_time(t, epoch)
        times = epoch + t

        if np.any(t.sec < 0):
            t_order = np.argsort(-t.sec)
        else:
            t_order = np.argsort(t.sec)

        if isinstance(state0, pyorb.Orbit):
            state0_cart = np.squeeze(state0.cartesian)
        else:
            state0_cart = state0

        if len(state0_cart.shape) > 1:
            if state0_cart.shape[1] > 1:
                N_testparticle = state0_cart.shape[1]
            else:
                N_testparticle = 1
        else:
            N_testparticle = 1

        if not (np.all(t.sec <= 0) or np.all(t.sec >= 0)):
            states = np.empty((6, len(t), N_testparticle), dtype=np.float64)
            if N_testparticle == 1:
                states.shape = states.shape[:2]

            ret_backward = self.propagate(t[t.sec < 0], state0, epoch, **kwargs)
            ret_forward = self.propagate(t[t.sec >= 0], state0, epoch, **kwargs)

            if self.settings["save_massive_states"]:
                massive_states = np.empty((6, len(t), self.N_massive), dtype=np.float64)

                states_b, massive_b = ret_backward
                states_f, massive_f = ret_forward

                massive_states[:, t.sec < 0, :] = massive_b
                massive_states[:, t.sec >= 0, :] = massive_f
            else:
                states_b = ret_backward
                states_f = ret_forward

            states[:, t.sec < 0, ...] = states_b
            states[:, t.sec >= 0, ...] = states_f

            if self.settings["save_massive_states"]:
                return states, massive_states
            else:
                return states

        if np.all(t.sec <= 0):
            backwards_integration = True
            t = -t
        else:
            backwards_integration = False

        t_restore = np.argsort(t_order)
        t = t[t_order]
        times = times[t_order]

        init_massive_states = kwargs.get("massive_states", None)

        self._setup_sim(epoch, init_massive_states=init_massive_states)

        m = kwargs.get("m", np.zeros((N_testparticle,), dtype=np.float64))
        if isinstance(m, float) or isinstance(m, int):
            m = np.zeros((N_testparticle,), dtype=np.float64) * m

        if frames.is_geocentric(self.settings["in_frame"]):
            earth_state = self._get_earth_state()

            state0_cart = frames.convert(
                epoch,
                state0_cart,
                in_frame=self.settings["in_frame"],
                out_frame=self.geo_internal_frame,
            )

            if len(state0_cart.shape) > 1:
                state0_cart = state0_cart + earth_state[:, None]
            else:
                state0_cart = state0_cart + earth_state
        else:
            state0_cart = frames.convert(
                epoch,
                state0_cart,
                in_frame=self.settings["in_frame"],
                out_frame=self.internal_frame,
            )

        if len(state0_cart.shape) > 1:
            for ni in range(N_testparticle):
                self._add_state(state0_cart[:, ni], m[ni])
        else:
            self._add_state(state0_cart, m[0])

        self.sim.move_to_com()

        # Fix for backwards integration, Rebound cannot handle negative times as it is trivial to reverse time
        if backwards_integration:
            for p_ in self.sim.particles:
                p_.vx = -p_.vx
                p_.vy = -p_.vy
                p_.vz = -p_.vz

        if self.settings["save_massive_states"]:
            massive_states = np.empty((6, len(t), self.N_massive), dtype=np.float64)
        else:
            massive_states = None

        states = np.empty((6, len(t), N_testparticle), dtype=np.float64)
        end_ind = len(t)

        if self.settings["tqdm"]:
            pbar = tqdm(total=len(t))

        for ti in range(len(t)):
            self.sim.integrate(t[ti].sec)

            if self.settings["tqdm"]:
                pbar.update(1)

            massive_states, states = self._put_simulation_state(massive_states, states, ti)

            if frames.is_geocentric(self.settings["out_frame"]):
                earth_state = self._get_earth_state()
                states[:, ti, :] -= earth_state[:, None]
                if self.settings["save_massive_states"]:
                    massive_states[:, ti, :] -= earth_state[:, None]
            else:
                sun_state = self._get_helio_state()
                states[:, ti, :] -= sun_state[:, None]
                if self.settings["save_massive_states"]:
                    massive_states[:, ti, :] -= sun_state[:, None]

            if (
                self.settings["termination_check"]
                and ti % self.settings["termination_check_interval"] == 0
            ):
                if backwards_integration:
                    t__ = -t[ti]
                else:
                    t__ = t[ti]
                if self.termination_check(t__, ti, massive_states, states):
                    end_ind = ti + 1
                    break

        if self.settings["tqdm"]:
            pbar.close()

        if backwards_integration:
            states[3:, :, :] = -states[3:, :, :]
            t = -t

        t = t[0:end_ind]
        times = times[0:end_ind]
        t_restore = t_restore[0:end_ind]
        states = states[:, 0:end_ind, :]

        if frames.is_geocentric(self.settings["out_frame"]):
            int_frame_ = self.geo_internal_frame
        else:
            int_frame_ = self.internal_frame

        for ni in range(N_testparticle):
            states[:, :, ni] = frames.convert(
                times,
                states[:, :, ni],
                in_frame=int_frame_,
                out_frame=self.settings["out_frame"],
            )

        states = states[:, t_restore, :]
        if N_testparticle == 1:
            states.shape = states.shape[:2]

        if self.settings["save_massive_states"]:
            if backwards_integration:
                massive_states[3:, :, :] = -massive_states[3:, :, :]
            massive_states = massive_states[:, 0:end_ind, :]

            for ni in range(self.N_massive):
                massive_states[:, :, ni] = frames.convert(
                    times,
                    massive_states[:, :, ni],
                    in_frame=int_frame_,
                    out_frame=self.settings["out_frame"],
                )

            massive_states = massive_states[:, t_restore, :]

            return states, massive_states
        else:
            return states



from __future__ import annotations
import logging, typing as t
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
import xarray as xr
from sorts import radar, scheduling
from sorts.utils import to_datetime64_us, to_timedelta64_us
from sorts.space_object import SpaceObject
from sorts.radar import Station
from sorts.types import (
    EcefStates,
    Datetime64_us,
    Timedelta64_us,
    EnuCoordinates,
    Datetime_Like,
)
from sorts.simulation.types import Passage
from .controller_base import ControllerBase
from sorts.interpolation import Interpolator

logger = logging.getLogger(__name__)


# TODO: this is probably a very general thing that can be just a
# standard type in `types`, we could then pass these around to e.g. the
# interpolator
# @dataclass
# class ControllerState:
#     spobj_time: npt.NDArray[Datetime64_us]
#     spobj_states: EcefStates
#
#     @classmethod
#     def empty(cls) -> t.Self:
#         return cls(
#             spobj_time=np.empty(0, dtype="datetime64[us]"),
#             spobj_states=np.empty((6, 0), dtype=np.float64),
#         )


class SparseTrackerController(ControllerBase):
    """
    Generate pointing schedule that tracks a space object.

    - The preferred way to create instances of this class is via its class methods (e.g. `TrackerController.from_space_object`).
    - This class serve as a frontend to the `State` type in this module
    """

    def __init__(
        self,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        exp_detail: scheduling.ExperimentDetail,
        space_object: SpaceObject,
        epoch: Datetime64_us,
        station_id_pairs: list[tuple[radar.StationId, radar.StationId]],
        points_per_passage: int,
        interpolator: Interpolator,
    ):
        """
        NOTE: This is intended as an internal constructor, please use the constructor methods to create instances.
        """

        self.tx_station = tx_station
        self.rx_stations = rx_stations
        self.exp_detail = exp_detail
        self.space_object = space_object
        self.epoch = epoch
        self.station_id_pairs = station_id_pairs
        self.points_per_passage = points_per_passage

        self.interpolator = interpolator

    @classmethod
    def from_space_object(
        cls,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        exp_detail: scheduling.ExperimentDetail,
        space_object: SpaceObject,
        epoch: Datetime_Like,
        points_per_passage: int,
        interpolator: Interpolator,
    ) -> t.Self:
        """A constructor method"""

        stn_pairs = [(tx_station.uid, rx_station.uid) for rx_station in rx_stations]

        ctrl = cls(
            tx_station=tx_station,
            rx_stations=rx_stations,
            exp_detail=exp_detail,
            space_object=space_object,
            epoch=to_datetime64_us(epoch),
            station_id_pairs=stn_pairs,
            points_per_passage=points_per_passage,
            interpolator=interpolator,
        )

        return ctrl

    def get_experiment_detail(self) -> scheduling.ExperimentDetail:
        return self.exp_detail

    def get_experiment_id_station_id_pairs_map(self) -> scheduling.ExperimentIdStationIdPairsMap:
        return {self.exp_detail.id: self.station_id_pairs}

    def get_station_map(self) -> dict[radar.StationId, radar.Station]:
        stn_map: dict[radar.StationId, radar.Station] = {}

        stn_map[self.tx_station.uid] = self.tx_station
        stn_map.update(list([(stn.uid, stn) for stn in self.rx_stations]))

        return stn_map

    def generate(self, passages_of_spobj: list[Passage]) -> scheduling.Schedule:
        """Generate the schedules."""
        # early return for empty case
        if len(passages_of_spobj) == 0:
            return scheduling.empty()

        observation_times_relative = []
        observation_times = []
        min_time_needed = (
            self.points_per_passage * self.exp_detail.slice_duration / np.timedelta64(1, "s")
        )
        for ps in passages_of_spobj:
            pstart_time, pend_time = ps.time_range
            t0 = (pstart_time - ps.epoch) / np.timedelta64(1, "s")
            passage_time = (pend_time - pstart_time) / np.timedelta64(1, "s")
            if passage_time <= min_time_needed:
                continue

            relative_time_sampling = np.linspace(
                0.0, passage_time, num=self.points_per_passage + 2, endpoint=True
            )
            relative_time_sampling = relative_time_sampling[1:-1]

            observation_times_relative.append(t0 + relative_time_sampling)
            rel_us = (relative_time_sampling.copy() * 1e6).astype("timedelta64[us]")

            observation_times.append(pstart_time + rel_us)

        # todo: this can be cleaned up quite a lot i feel like
        tx_sch_time_rel = np.concatenate(observation_times_relative)
        tx_sch_time = np.concatenate(observation_times)
        tx_sch_len = len(tx_sch_time)
        tx_pointings: EnuCoordinates = self.tx_station.enu(
            self.interpolator.get_state(tx_sch_time_rel)[:3, :]
        )
        tx_pointings = tx_pointings / np.linalg.norm(tx_pointings, axis=0)

        tx_sch = scheduling.from_ndarrays(
            start_time=tx_sch_time,
            end_time=tx_sch_time + self.exp_detail.slice_duration,
            exp_num=np.full(tx_sch_len, self.exp_detail.id, dtype=np.int16),
            stn_num=np.full(tx_sch_len, self.tx_station.uid, dtype=np.int16),
            simult_num=np.full(tx_sch_len, 0, dtype=np.int16),
            pointing=tx_pointings,
        )

        rx_schs: list[scheduling.Schedule] = []
        for rx_stn in self.rx_stations:
            rx_pointings: EnuCoordinates = rx_stn.enu(
                self.interpolator.get_state(tx_sch_time_rel)[:3, :]
            )
            rx_pointings = rx_pointings / np.linalg.norm(rx_pointings, axis=0)

            rx_schs.append(
                scheduling.from_ndarrays(
                    start_time=tx_sch_time,
                    end_time=tx_sch_time + self.exp_detail.slice_duration,
                    exp_num=np.full(tx_sch_len, self.exp_detail.id, dtype=np.int16),
                    stn_num=np.full(tx_sch_len, rx_stn.uid, dtype=np.int16),
                    simult_num=np.full(tx_sch_len, 0, dtype=np.int16),
                    pointing=rx_pointings,
                )
            )

        resultant_sch = xr.concat([tx_sch, *rx_schs], dim=scheduling._K.multi_index)
        resultant_sch = resultant_sch.sortby(scheduling._K.start_time)
        output = resultant_sch

        return output



#!/usr/bin/env python

"""This module is used to define the radar controller

"""
# Python standard import
from abc import ABC, abstractmethod

# Third party import
import numpy as np

# Local import


class RadarController(ABC):
    """A radar controller."""

    META_FIELDS = [
        "controller_type",
        "t_slice",
    ]

    def __init__(self, radar, t=None, t0=0.0, t_slice=None, meta=None):
        self.radar = radar
        self.t = t
        self.t0 = t0
        self.t_slice = t_slice

        self.meta = dict()
        if meta is not None:
            self.meta.update(meta)
        for key in self.META_FIELDS:
            if key not in self.meta:
                self.meta[key] = None

    def run(self):
        return self(self.t - self.t0)

    def toggle_stations(self, t, radar):
        if self.t is None:
            RadarController.turn_on(radar)
        else:
            dt = t - self.t
            check = np.logical_and(dt >= 0, dt < self.t_slice)
            if np.any(check):
                RadarController.turn_on(radar)
            else:
                RadarController.turn_off(radar)

    def default_meta(self):
        """This is used to generate meta data on the fly, rather then the static data that can be set in the `self.meta`."""
        meta = dict()
        meta.update(self.meta)
        meta["controller_type"] = self.__class__
        meta["t_slice"] = self.t_slice
        return meta

    @abstractmethod
    def generator(self, t, **kwargs):
        """This will configure the radar system and return a pointer to the contained radar system instance with the correct configuration.
        It should always assume the input `t` is an iterable and use `yield` to return `radar, meta`. The `meta` variable should be
        a dict with the fields defined in `META_FIELDS`

        **NOTE:** This is NOT guaranteed to return a copy of the radar system, however, the subclass should implement this as a option.
        """
        pass

    def __call__(self, t, **kwargs):
        if isinstance(t, float) or isinstance(t, int):
            ret = list(self.generator([t], **kwargs))[0]
        else:
            if len(t) > 0:
                ret = self.generator(t, **kwargs)
            else:
                ret = []

        return ret

    @staticmethod
    def _point_station(station, ecef):
        if len(ecef.shape) > 1:
            k = station.point_ecef(ecef - station.ecef[:, None])
        else:
            k = station.point_ecef(ecef - station.ecef)

        # pointing turns on station
        station.enabled = True

        # error check pointing
        keep = k[2, ...] >= np.sin(np.radians(station.min_elevation))
        if len(ecef.shape) > 1:
            if not np.any(keep):
                station.enabled = False
            else:
                new_k = k[:, keep]
                station.point(new_k)
        else:
            if not keep:
                station.enabled = False

    @staticmethod
    def point_rx_ecef(radar, ecef):
        """Point all rx sites into the direction of given ECEF coordinate, relative Earth Center."""
        for rx in radar.rx:
            RadarController._point_station(rx, ecef)

    @staticmethod
    def point_tx_ecef(radar, ecef):
        """Point all tx sites into the direction of given ECEF coordinate, relative Earth Center."""
        for tx in radar.tx:
            RadarController._point_station(tx, ecef)

    @staticmethod
    def point_ecef(radar, ecef):
        """Point all sites into the direction of given ECEF coordinate, relative Earth Center."""
        RadarController.point_tx_ecef(radar, ecef)
        RadarController.point_rx_ecef(radar, ecef)

    @staticmethod
    def coh_integration(radar, dwell):
        """Set the coherent integration settings based on the dwell time."""
        for tx in radar.tx:
            tx.n_ipp = int(dwell / tx.ipp)
            tx.coh_int_bandwidth = 1.0 / (tx.pulse_length * tx.n_ipp)

    @staticmethod
    def point(radar, enu):
        """Point all sites into the direction of a given East, North, Up (ENU) local coordinate system."""
        for tx in radar.tx:
            tx.point(enu)
        for rx in radar.rx:
            rx.point(enu)

    @staticmethod
    def turn_off(radar):
        for st in radar.tx + radar.rx:
            st.enabled = False

    @staticmethod
    def turn_on(radar):
        for st in radar.tx + radar.rx:
            st.enabled = True



"""
Shared types in this package.

(Types might live in their own module instead of here if it improves readability,
and the imports can be worked around, e.g, by `if t.TYPE_CHECKING`)
"""
from copy import deepcopy
from dataclasses import dataclass, fields
import typing as t
import numpy as np
import numpy.typing as npt
from datetime import datetime, timedelta
from astropy.time import Time, TimeDelta


Datetime64_us = np.datetime64
"`numpy` `datetime64` in `us` resolution"

Timedelta64_us = np.timedelta64
"`numpy` `timedelta64` in `us` resolution"

Datetime_Like = datetime | Time | Datetime64_us | str
"""
One of:
- python builtin `datetime`,
- astropy `Time`
- numpy `timedelta64`
- ISO 8601 date or datetime string supported by numpy
"""

Timedelta_Like = timedelta | TimeDelta | Timedelta64_us | int
"""
One of:
- python builtin `timedelta`,
- astropy `TimeDelta`
- numpy `timedelta64`
- `int`, the length of the duration in `us` resolution
"""

TimeRange_us = tuple[Datetime64_us, Datetime64_us]
"""The start time and end time of the passage, a right-open interval"""

Float_as_sec = float
"`float` as second"

Float_as_m = float
"`float` as meters"

Float_as_deg = float
"`float` as angle in degrees"

Float64_as_sec = np.float64
"`numpy` `float64` as seconds"

Float64_as_m = np.float64
"`numpy` `float64` as meters"

Float64_as_deg = np.float64
"`numpy` `float64` as angle in degrees"

Float64_as_rad = np.float64
"`numpy` `float64` as angle in radians"

# TODO: make sure these comments make sense
NDArray_3x1 = npt.NDArray
"(3,) shaped ndarray (i.e. a single 3D vector)"

NDArray_6x1 = npt.NDArray
"(6,) shaped ndarray (i.e. a single 6D vector)"

NDArray_N = npt.NDArray
"(n,) shaped ndarray (i.e. a single `n`D vector)"

NDArray_6 = npt.NDArray
"(6,) shaped ndarray (i.e. a single `6`D vector)"

NDArray_3xN = npt.NDArray
"(3,n) shaped ndarray (i.e. `3` rows and `n` columns)"

NDArray_6xN = npt.NDArray
"(6,n) shaped ndarray (i.e. `6` 1D vectors of length `n`)"

NDArray_Nx3 = npt.NDArray
"(n, 3) shaped ndarray (i.e. `n` 3D vectors)"

NDArray_Nx6 = npt.NDArray
"(n, 6) shaped ndarray (i.e. `n` 6D vectors)"

EnuCoordinate = NDArray_3x1[np.float64]
"ENU cartesian coordinate, a (3,) shaped ndarray of `float64` (i.e. a single 3D vector)"

EnuCoordinates = NDArray_3xN[np.float64]
"ENU cartesian coordinates, a `(3,n)` ndarray of `float64`"

AzelCoordinates_DegM = NDArray_3xN[np.float64]
"""
`(Azimuth, Elevation)` spherical coordinates in degrees; a `(2,n)` ndarray of `float64`

- Azimuth should be in [-180, 180)
- Elevation should be in [0, 90]
"""

AzelrCoordinates_DegM = NDArray_3xN[np.float64]
"""
`(Azimuth, Elevation, Range)` spherical coordinates in degrees and meters; a `(3,n)` ndarray of `float64`

- Azimuth should be in [-180, 180)
- Elevation should be in [0, 90]
"""

GeodeticCoordinates_DegM = NDArray_3xN[np.float64]
"""
`(Latitude, Longitude, Height)` geodetic coordinates in degrees and meters; a `(3,n)` ndarray of `float64`

- Latitude should be in [-90, 90]
- Longitude should be in [-180, 180)
"""

EcefCoordinates = NDArray_3xN[np.float64]
"ECEF cartesian coordinates, a `(3,n)` ndarray of `float64`"

EcefStates = NDArray_6xN[np.float64]
"ECEF states in cartesian coordinate, a `(6,n)` ndarray of `float64`, usually used for space objects"


class TxRxTuple[TxType, RxType](t.NamedTuple):
    tx: TxType
    rx: RxType


type Tuple_3[T] = tuple[T, T, T]
type Tuple_6[T] = tuple[T, T, T, T, T, T]
type Tuple_7[T] = tuple[T, T, T, T, T, T, T]
type SpaceObjectId = int

IndexLike = int | list[int] | tuple[int] | NDArray_N | slice | np.integer

S = t.TypeVar("S", bound="Settings")

Frames = t.Literal[
    "TEME",
    "ITRS",
    "ITRF",
    "ICRS",
    "ICRF",
    "GCRS",
    "GCRF",
    "HCRS",
    "HCRF",
]
GravModels = t.Literal["WGS84"]
StateType = t.Literal["kepler", "cartesian"]
AnomalyType = t.Literal["mean", "eccentric", "true"]


@dataclass
class Settings:
    def copy(self: S) -> S:
        kwargs = {key: deepcopy(getattr(self, key)) for key in self.keys}
        return self.__class__(**kwargs)

    @property
    def keys(self) -> list[str]:
        return [key.name for key in fields(self)]



from .types import Passage, SpaceObjectJacobianTuple
from . import types, funcs, stx_mrx_simulation

# TODO: remove `Spec` from here; at the use site, import it from the re-exported `stx_mrx_simulation` module instead
from .stx_mrx_simulation import StxMrxSimulation



#!/usr/bin/env python

""" """

import pyant

from .radars import radar_generator
from ..radar import Radar
from ..tx_rx import TX, RX
from ... import frames


@radar_generator("eiscat_uhf", "measured")
def gen_eiscat_uhf():
    lat_tro = 69.0 + frames.arctime_to_degrees(35, 11)
    lon_tro = 19.0 + frames.arctime_to_degrees(13, 38)
    alt_tro = 86.0

    lat_krn = 67.0 + frames.arctime_to_degrees(51, 38)
    lon_krn = 20.0 + frames.arctime_to_degrees(26, 7)
    alt_krn = 418.0

    lat_sod = 67.0 + frames.arctime_to_degrees(21, 49)
    lon_sod = 26.0 + frames.arctime_to_degrees(37, 37)
    alt_sod = 197.0

    tx_beam = pyant.beam_of_radar("eiscat_uhf", "measured")
    rx_beam = pyant.beam_of_radar("eiscat_uhf", "measured")

    tx = [
        TX(
            lat=lat_tro,
            lon=lon_tro,
            alt=alt_tro,
            min_elevation=30,
            beam=tx_beam,
            power=1.6e6,
            bandwidth=1e6,
            duty_cycle=0.125,
            pulse_length=30.0 * 64.0 * 1e-6,
            ipp=20e-3,
            n_ipp=10.0,
        )
    ]

    rx = []

    rx += [
        RX(
            lat=lat_tro,
            lon=lon_tro,
            alt=alt_tro,
            min_elevation=30,
            noise=100,
            beam=rx_beam.copy(),
        )
    ]
    rx += [
        RX(
            lat=lat_krn,
            lon=lon_krn,
            alt=alt_krn,
            min_elevation=30,
            noise=100,
            beam=rx_beam.copy(),
        )
    ]
    rx += [
        RX(
            lat=lat_sod,
            lon=lon_sod,
            alt=alt_sod,
            min_elevation=30,
            noise=100,
            beam=rx_beam.copy(),
        )
    ]

    uhf = Radar(
        tx,
        rx,
        min_SNRdb=10.0,
        joint_stations=[(0, 0)],
    )
    return uhf



import typing as t
import numpy as np
import numpy.typing as npt
import bokeh.plotting as bp
import bokeh.models as bokeh_models
from sorts import scheduling
from sorts.types import Datetime_Like, Datetime64_us
from sorts.utils import to_datetime64_us
from sorts.scheduling import Schedule


def _schedule_plot_from_cds(
    source: bokeh_models.ColumnarDataSource,
    start_time: Datetime64_us,
    end_time: Datetime64_us,
    y_range: t.Sequence[str] | npt.NDArray[np.str_],
):
    """An internal ver of `schedule_plot` that takes a bokeh `ColumnDataSource`"""

    _SK = scheduling._K

    bar = bp.figure(
        y_range=y_range,  # type: ignore
        x_axis_type="datetime",
        x_axis_location="above",
        width=800,
        height=300,
    )
    bar.add_tools(bokeh_models.HoverTool())

    bar.hbar(
        y=_SK.exp_num,
        left=_SK.start_time,
        right=_SK.end_time,
        source=source,
    )
    bar.x_range.range_padding = 0  # type: ignore
    bar_xpan_tool = bokeh_models.PanTool(dimensions="width")
    bar_xwheel_zoom_tool = bokeh_models.WheelZoomTool(dimensions="width")
    bar.add_tools(bar_xpan_tool)
    bar.add_tools(bar_xwheel_zoom_tool)
    bar.toolbar.active_drag = bar_xpan_tool
    bar.toolbar.active_scroll = bar_xwheel_zoom_tool

    minimap = bp.figure(
        title="Drag the middle and edges of the selection box to change the range above",
        height=130,
        width=800,
        x_axis_type="datetime",
        y_axis_type=None,
        tools="",
        toolbar_location=None,
    )
    minimap.x_range.range_padding = 0.05  # type: ignore
    minimap.x_range.bounds = "auto"  # type: ignore

    # NOTE: a dummy line is plotted; select tool doesn't work well without any data plotted
    minimap.line(
        x=[start_time, end_time],  # type: ignore
        y=[0, 0],
    )
    minimap_range_tool = bokeh_models.RangeTool(x_range=bar.x_range, start_gesture="pan")
    minimap.add_tools(minimap_range_tool)

    plot = bp.column(bar, minimap)

    return plot, bar, minimap


# TODO: add time based binning and aggregation
def schedule_plot(
    sch: Schedule,
    start_time: Datetime_Like | None = None,
    end_time: Datetime_Like | None = None,
):
    """
    Note:
    Plotting the full schedule can be computationally demanding and lead to application crashes.
    Limiting the plot range by `start_time` and `end_time` param is recommended.

    Without aggregations, a good starting point is a 5 minutes time range.
    """

    _SK = scheduling._K

    df = scheduling.to_dataframe(sch)

    start_time_: Datetime64_us = (
        to_datetime64_us(start_time) if start_time is not None else df[_SK.start_time].min()
    )
    end_time_: Datetime64_us = (
        to_datetime64_us(end_time) if end_time is not None else df[_SK.end_time].max()
    )

    df = df[(df[_SK.start_time] >= start_time_) & (df[_SK.end_time] <= end_time_)]
    # bokeh requires str type for categorical axis
    df[_SK.exp_num] = df[_SK.exp_num].astype(str)

    plot, *_ = _schedule_plot_from_cds(
        source=bokeh_models.ColumnDataSource(df),
        start_time=start_time_,
        end_time=end_time_,
        y_range=df[_SK.exp_num].unique(),
    )

    return plot



#!/usr/bin/env python
import sys
import re
import getpass
import argparse
from datetime import datetime, timedelta
import subprocess

try:
    import spacetrack
except ImportError:
    spacetrack = None

from .commands import add_command

_iso_fmt = "%Y-%m-%d"
_td_regx = re.compile(
    r"^((?P<days>[\.\d]+?)d)? *((?P<hours>[\.\d]+?)h)? "
    + r"*((?P<minutes>[\.\d]+?)m)? *((?P<seconds>[\.\d]+?)s)?$"
)


def parse_timedelta(time_str):
    """
    Parse a time string e.g. (2h13m) into a timedelta object.

    Modified from virhilo's answer at https://stackoverflow.com/a/4628148/851699

    :param time_str: A string identifying a duration.  (eg. 2h13m)
    :return datetime.timedelta: A datetime.timedelta object
    """
    parts = _td_regx.match(time_str)
    assert parts is not None, (
        f"Could not parse any time information from '{time_str}'.  "
        + "Examples of valid strings: '8h', '2d8h5m20s', '2m4s'"
    )
    time_params = {name: float(param) for name, param in parts.groupdict().items() if param}
    return timedelta(**time_params)


def main(args):
    assert spacetrack is not None, "spacetrack package not installed, run `pip install spacetrack`"
    # end date argument can be ISO format datetime or 'now'
    if args.end_date == "now":
        dt1 = datetime.now()
    else:
        dt1 = datetime.strptime(args.end_date, _iso_fmt)

    # start date argument can be absolute or timedelta
    try:
        dt0 = dt1 - parse_timedelta(args.start_date)
    except AssertionError:
        dt0 = datetime.strptime(args.start_date, _iso_fmt)

    if args.output is not sys.stdout:
        print(f"Getting TLEs for the range [{dt0} -> {dt1}]")
        print(f"Output to {args.output.name}")

    drange = spacetrack.operators.inclusive_range(dt0, dt1)
    kwargs = {}

    if args.name is not None:
        name_op = spacetrack.operators.like(args.name)
        kwargs["object_name"] = name_op
        kwargs["epoch"] = drange
    else:
        kwargs["publish_epoch"] = drange

    if args.secret_tool_key is not None:
        res = subprocess.run(
            ["secret-tool", "lookup", "username"] + args.secret_tool_key,
            capture_output=True,
            text=True,
        )
        user = res.stdout
        res = subprocess.run(
            ["secret-tool", "lookup", "password"] + args.secret_tool_key,
            capture_output=True,
            text=True,
        )
        passwd = res.stdout
    elif args.credentials is not None:
        raise NotImplementedError("Add input of username/password from file")
    else:
        user = input("Username for space-track.org:")
        passwd = getpass.getpass("Password for " + user + ":")

    st = spacetrack.SpaceTrackClient(user, passwd)

    if args.name is not None:
        print('Using CLASS "tle"...')
        lines = st.tle(iter_lines=True, orderby="TLE_LINE1", format="tle", **kwargs)
    else:
        print('Using CLASS "tle_publish"...')
        lines = st.tle_publish(iter_lines=True, orderby="TLE_LINE1", format="tle", **kwargs)
    lineno = 0
    for line in lines:
        args.output.write(line + "\n")
        lineno += 1

    if args.output is not sys.stdout:
        print(f"Wrote {lineno} lines to {args.output.name}")


def parser_build(parser):
    parser.add_argument(
        "start_date",
        type=str,
        nargs="?",
        default="7d",
        help='Start date of snapshot [ISO] or timedelta ("24h", "12d", etc)',
    )
    parser.add_argument(
        "end_date", type=str, nargs="?", default="now", help="End date of snapshot [ISO]"
    )
    parser.add_argument("output", nargs="?", type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("--secret-tool-key", "-k", nargs=1)
    parser.add_argument(
        "--credentials",
        "-c",
        nargs=1,
        help="File containing username and password for space-track.org",
    )
    parser.add_argument(
        "--name", "-n", default=None, help='Name of the object to match with the "like" operator'
    )

    return parser


add_command(
    name="spacetrack",
    function=main,
    parser_build=parser_build,
    command_help="Download tle snapshot from space-track",
)



#!/usr/bin/env python

"""Module for using REBOUND to calculate pre-encounter states of objects that make close-encounters/collisions with Earth.


"""

import sys
import os
import time
import glob

from tqdm import tqdm
import numpy as np
import scipy
import h5py
from astropy.time import Time, TimeDelta

import pyorb

from .propagator import Rebound


def distance_termination(dAU):
    def distance_termination_method(self, t, step_index, massive_states, particle_states):
        e_state = massive_states[:3, step_index, self._earth_ind]
        d_earth = (
            np.linalg.norm(particle_states[:3, step_index, :] - e_state[:, None], axis=0) / pyorb.AU
        )
        return np.all(d_earth > dAU)

    return distance_termination_method


def propagate_pre_encounter(
    state,
    epoch,
    in_frame,
    out_frame,
    termination_check,
    kernel,
    dt=10.0,
    max_t=10 * 24 * 3600.0,
    settings=None,
):
    """Propagates a state from the states backwards in time until the termination_check is true.

    #TODO: implement no saving of intermediate states for large particle runs
    """
    t = -np.arange(0, max_t, dt, dtype=np.float64)

    class TerminatedRebound(Rebound):
        pass

    TerminatedRebound.termination_check = termination_check

    reb_settings = dict(
        in_frame=in_frame,
        out_frame=out_frame,
        time_step=dt,  # s
        termination_check=True,
        save_massive_states=True,
    )
    if settings is not None:
        settings.update(reb_settings)
    else:
        settings = reb_settings

    prop = TerminatedRebound(
        kernel=kernel,
        settings=settings,
    )

    states, massive_states = prop.propagate(t, state, epoch)

    t = t[: states.shape[1]]

    return states, massive_states, t



#!/usr/bin/env python

"""
"""

import pyant
import numpy as np

from ..radar import Radar
from ..tx_rx import TX, RX


def gen_mock():
    class Omni(pyant.Beam):
        def copy(self):
            ret = Omni(
                azimuth=self.azimuth,
                elevation=self.elevation,
                frequency=self.frequency,
                radians=self.radians,
            )
            return ret

        def gain(self, k, polarization=None, ind=None):
            if len(k.shape) == 1:
                return 1.0
            else:
                return np.ones((k.shape[1],), dtype=k.dtype)

    tx_beam = Omni(azimuth=0.0, elevation=90.0, frequency=100e6)
    rx_beam = Omni(azimuth=0.0, elevation=90.0, frequency=100e6)

    tx = [
        TX(
            lat=90,
            lon=0,
            alt=0,
            min_elevation=0,
            beam=tx_beam,
            power=1.6e6,
            bandwidth=1e6,
            duty_cycle=0.25,
            pulse_length=30.0 * 64.0 * 1e-6,
            ipp=20e-3,
            n_ipp=10.0,
        )
    ]
    rx = [
        RX(
            lat=90,
            lon=0,
            alt=0,
            min_elevation=0,
            noise=100,
            beam=rx_beam,
        )
    ]
    mock = Radar(
        tx,
        rx,
        min_SNRdb=10.0,
    )
    return mock



#!/usr/bin/env python

"""Radar scan plot functions"""

# Python standard import
from itertools import cycle

# Third party import
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D


# Local import
from . import general
from ..controller_v1 import Tracker

prop_cycle = plt.rcParams["axes.prop_cycle"]


def unit_vector(v):
    return v / np.sqrt(np.sum(v * v, axis=-1))[:, None]


def observed_parameters(data_list, passes=None, snrdb_lim=10.0, axes=None, sort=True, **kwargs):
    """Observed parameters for one RX station."""

    color = cycle(prop_cycle.by_key()["color"])

    if axes is None:
        if "figsize" in kwargs:
            fig = plt.figure(figsize=kwargs["figsize"])
        else:
            fig = plt.figure()
        axes = np.array(
            [
                [
                    fig.add_subplot(221),
                    fig.add_subplot(222),
                ],
                [
                    fig.add_subplot(223),
                    fig.add_subplot(224),
                ],
            ]
        )
    else:
        fig = None

    if kwargs.get("plot_origin", True):
        axes[0][0].plot([0], [0], [0], ".k")

    line_all = kwargs.get("linestyle_all", "-")
    line_det = kwargs.get("linestyle_detect", ".r")
    fontsize = kwargs.get("fontsize", None)

    text_spec = dict(fontsize=fontsize)

    text_spec = {key: val for key, val in text_spec.items() if val is not None}

    time_unit = kwargs.get("time_unit", "min").strip().lower()
    if time_unit in ["s", "sec", "seconds"]:
        _tt = 1.0
        t_label = "[s]"
    elif time_unit in ["m", "min", "minutes"]:
        _tt = 60.0
        t_label = "[min]"
    elif time_unit in ["h", "hour", "hours"]:
        _tt = 3600.0
        t_label = "[h]"
    elif time_unit in ["d", "days"]:
        _tt = 3600.0 * 24
        t_label = "[d]"
    else:
        raise TypeError(f'time unit "{time_unit}" not found')

    max_snr = [0]
    for pi, dat in enumerate(data_list):
        if dat is None:
            continue
        cc = next(color)

        if passes and passes[pi] is not None:
            # If we have passes, use those for nice plotting
            tvec = passes[pi].t
            rx_k = unit_vector(passes[pi].enu[1][:3].T).T
            rng = passes[pi].range()
            rng = rng[0] + rng[1]
            rrt = passes[pi].range_rate()
            rrt = rrt[0] + rrt[1]
        else:
            tvec = dat["t"]
            rx_k = dat["rx_k"]
            rng = dat["range"]
            rrt = dat["range_rate"]

        if sort:
            inds = np.argsort(tvec)
            t_ind = np.argsort(dat["t"])
        else:
            inds = np.arange(len(tvec))
            t_ind = np.arange(len(dat["t"]))

        SNRdB = 10 * np.log10(dat["snr"])
        det_inds = SNRdB > snrdb_lim

        # axes[0][0].plot(dat['rx_k'][0,inds], dat['rx_k'][1,inds], line_all)
        axes[0][0].plot(rx_k[0, inds], rx_k[1, inds], line_all, color=cc, label=f"Pass{pi}")
        axes[0][0].plot(dat["rx_k"][0, det_inds], dat["rx_k"][1, det_inds], line_det)

        # axes[0][1].plot(dat['t'][inds]/_tt, dat['range'][inds]*1e-3, line_all, label=f'Pass{pi}')
        # axes[1][0].plot(dat['t'][inds]/_tt, dat['range_rate'][inds]*1e-3, line_all)
        axes[0][1].plot(tvec[inds] / _tt, rng[inds] * 1e-3, line_all, color=cc, label=f"Pass{pi}")
        axes[1][0].plot(tvec[inds] / _tt, rrt[inds] * 1e-3, line_all, color=cc, label=f"Pass{pi}")
        axes[1][1].plot(dat["t"][t_ind] / _tt, SNRdB[t_ind], line_all, color=cc, label=f"Pass{pi}")

        axes[0][1].plot(dat["t"][det_inds] / _tt, dat["range"][det_inds] * 1e-3, line_det)
        axes[1][0].plot(dat["t"][det_inds] / _tt, dat["range_rate"][det_inds] * 1e-3, line_det)
        axes[1][1].plot(dat["t"][det_inds] / _tt, SNRdB[det_inds], line_det)
        max_snr.append(SNRdB.max())

    if np.max(max_snr) < 10.0:
        max_snr = [10]
    axes[1][1].set_ylim([0, np.max(max_snr)])

    axes[0][1].set_xlabel(f"Time {t_label}", **text_spec)
    axes[1][0].set_xlabel(f"Time {t_label}", **text_spec)
    axes[1][1].set_xlabel(f"Time {t_label}", **text_spec)

    axes[0][1].set_ylabel("Two way range [km]", **text_spec)
    axes[1][0].set_ylabel("Two way range rate [km/s]", **text_spec)
    axes[1][1].set_ylabel("SNR [dB]", **text_spec)

    # axes[0][1].legend()

    dr = 1
    axes[0][0].set_xlim([-dr, dr])
    axes[0][0].set_ylim([-dr, dr])
    axes[0][0].set_aspect(1)

    axes[0][0].set_xlabel("East [k_x]", **text_spec)
    axes[0][0].set_ylabel("North [k_y]", **text_spec)

    if fig is not None:
        fig.tight_layout()

    return fig, axes


def schedule_pointing(
    scheduler,
    t0,
    t1,
    earth=True,
    earth_opts={},
    alpha=0.1,
    point_range=300e3,
    view_range=300e3,
    plot_rx=True,
    plot_tx=True,
    ax=None,
):
    """Plot the schedule generated by a `PointingSchedule` instance.

    :param bool earth: Plot the surface of the Earth.
    """

    if ax is None:
        fig = plt.figure(figsize=(15, 15))
        ax = fig.add_subplot(111, projection="3d")
        ax.grid(False)
    else:
        fig = None

    if earth:
        ax = general.grid_earth(ax, **earth_opts)

    for tx in scheduler.radar.tx:
        ax.plot([tx.ecef[0]], [tx.ecef[1]], [tx.ecef[2]], "or")
    for rx in scheduler.radar.rx:
        ax.plot([rx.ecef[0]], [rx.ecef[1]], [rx.ecef[2]], "og")

    def get_c_inds(c):
        if c.t_slice is not None:
            return np.logical_and(c.t + c.t_slice >= t0, c.t <= t1)
        else:
            return np.logical_and(c.t >= t0, c.t <= t1)

    # get sched data
    ctrls = scheduler.get_controllers()
    times = np.concatenate([c.t[get_c_inds(c)] for c in ctrls], axis=0)
    sched = scheduler.chain_generators([c(c.t[get_c_inds(c)] - c.t0) for c in ctrls])
    sched_data = scheduler.generate_schedule(times, sched)

    for ti, t in enumerate(sched_data["t"]):
        if plot_rx:
            for p, ecef in zip(sched_data["rx"][ti], sched_data["rx_pos"][ti]):
                if len(p.shape) == 1:
                    p.shape = (3, 1)

                for pi in range(p.shape[1]):
                    if "id" in sched_data["meta"][ti]:
                        ctrl = ctrls[sched_data["meta"][ti]["id"]]
                        if isinstance(ctrl, Tracker):
                            t_ind = np.argmin(np.abs(t - ctrl.t))
                            targetx = ctrl.ecefs[0, t_ind]
                            targety = ctrl.ecefs[1, t_ind]
                            targetz = ctrl.ecefs[2, t_ind]
                        else:
                            targetx = ecef[0] + p[0, pi] * point_range
                            targety = ecef[1] + p[1, pi] * point_range
                            targetz = ecef[2] + p[2, pi] * point_range
                    else:
                        targetx = ecef[0] + p[0, pi] * point_range
                        targety = ecef[1] + p[1, pi] * point_range
                        targetz = ecef[2] + p[2, pi] * point_range

                    ax.plot(
                        [ecef[0], targetx],
                        [ecef[1], targety],
                        [ecef[2], targetz],
                        "-g",
                        alpha=alpha,
                    )

        if plot_tx:
            for p, ecef in zip(sched_data["tx"][ti], sched_data["tx_pos"][ti]):
                if len(p.shape) == 1:
                    p.shape = (3, 1)

                for pi in range(p.shape[1]):
                    if "id" in sched_data["meta"][ti]:
                        ctrl = ctrls[sched_data["meta"][ti]["id"]]
                        if isinstance(ctrl, Tracker):
                            t_ind = np.argmin(np.abs(t - ctrl.t))
                            targetx = ctrl.ecefs[0, t_ind]
                            targety = ctrl.ecefs[1, t_ind]
                            targetz = ctrl.ecefs[2, t_ind]
                        else:
                            targetx = ecef[0] + p[0, pi] * point_range
                            targety = ecef[1] + p[1, pi] * point_range
                            targetz = ecef[2] + p[2, pi] * point_range
                    else:
                        targetx = ecef[0] + p[0, pi] * point_range
                        targety = ecef[1] + p[1, pi] * point_range
                        targetz = ecef[2] + p[2, pi] * point_range

                    ax.plot(
                        [ecef[0], targetx],
                        [ecef[1], targety],
                        [ecef[2], targetz],
                        "-r",
                        alpha=alpha,
                    )

    p0 = [tx.ecef for tx in scheduler.radar.tx]
    p0 = np.array(p0).mean(axis=0)

    ax.set_xlim(p0[0] - view_range, p0[0] + view_range)
    ax.set_ylim(p0[1] - view_range, p0[1] + view_range)
    ax.set_zlim(p0[2] - view_range, p0[2] + view_range)

    if fig is not None:
        fig.tight_layout()

    return fig, ax


def controller_slices(controllers, ax=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig = None

    times = {}
    max_t = 0
    for ctrl in controllers:
        var = ctrl.__class__.__name__
        if var not in times:
            times[var] = []

        if isinstance(ctrl.t_slice, float) or isinstance(ctrl.t_slice, int):
            t_slice = np.empty(ctrl.t.shape, dtype=np.float64)
            t_slice[:] = ctrl.t_slice
        else:
            t_slice = ctrl.t_slice

        for i in range(len(ctrl.t)):
            times[var].append((ctrl.t[i], t_slice[i]))

        if ctrl.t.max() > max_t:
            max_t = ctrl.t.max()

    ticks = []

    for var in times:
        ticks.append(var)
        ax.broken_barh(times[var], (len(ticks) * 10, 10))
        ax.plot(
            [x0 + 0.5 * x1 for x0, x1 in times[var]], [len(ticks) * 10 + 5 for x in times[var]], "."
        )

    ax.set_ylim(5, 15 + len(ticks) * 10)
    ax.set_xlim(0, max_t * 1.1)
    ax.set_xlabel("Time [s]")

    ax.set_yticks(list(range(15, 15 + 10 * len(ticks), 10)))
    ax.set_yticklabels(ticks)

    ax.grid(True)

    return fig, ax



#!/usr/bin/env python

"""General plotting functions

"""


# Third party import
import numpy as np

# Local import
from .. import constants
from .. import frames


def set_axes_equal(ax):
    """Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().

    Source: CC BY-SA 3.0 @ https://stackoverflow.com/a/31364297 by karlo

    """

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])


def grid_earth(
    ax,
    num_lat=25,
    num_lon=50,
    rotation=0,
    alpha=0.1,
    linewidth=1,
    res=100,
    color="black",
    hide_ax=True,
):
    """Add a 3d spherical grid to the given axis that represent the Earth."""
    lons = np.linspace(-180, 180, num_lon + 1) * np.pi / 180
    lons = lons[:-1]
    lats = np.linspace(-90, 90, num_lat) * np.pi / 180

    lonsl = np.linspace(-180, 180, res) * np.pi / 180
    latsl = np.linspace(-90, 90, res) * np.pi / 180

    rot = np.radians(rotation)

    r_e = constants.R_earth
    for lat in lats:
        x = r_e * np.cos(lonsl) * np.cos(lat)
        y = r_e * np.sin(lonsl) * np.cos(lat)
        z = r_e * np.ones(np.size(lonsl)) * np.sin(lat)

        x_ = x.copy()
        x = x_ * np.cos(rot) - y * np.sin(rot)
        y = x_ * np.sin(rot) + y * np.cos(rot)
        del x_

        ax.plot(x, y, z, alpha=alpha, linestyle="-", linewidth=linewidth, marker="", color=color)

    for lon in lons:
        x = r_e * np.cos(lon) * np.cos(latsl)
        y = r_e * np.sin(lon) * np.cos(latsl)
        z = r_e * np.sin(latsl)

        x_ = x.copy()
        x = x_ * np.cos(rot) - y * np.sin(rot)
        y = x_ * np.sin(rot) + y * np.cos(rot)
        del x_

        ax.plot(x, y, z, alpha=alpha, linestyle="-", linewidth=linewidth, marker="", color=color)

    if hide_ax:
        # Hide grid lines
        ax.grid(False)

        # Hide axes ticks
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])
        ax.set_axis_off()

    return ax


def transformed_grid_earth(
    ax,
    frame,
    time,
    num_lat=25,
    num_lon=50,
    alpha=0.1,
    linewidth=1,
    res=100,
    color="black",
    hide_ax=True,
):
    """Add a 3d spherical grid to the given axis that represent the Earth."""
    lons = np.linspace(-180, 180, num_lon + 1)
    lons = lons[:-1]
    lats = np.linspace(-90, 90, num_lat)

    lonsl = np.linspace(-180, 180, res)
    latsl = np.linspace(-90, 90, res)

    for lat in lats:
        pos = frames.geodetic_to_ITRS(np.ones_like(lonsl) * lat, lonsl, np.zeros_like(lonsl))
        pos = frames.convert(
            time,
            np.concatenate([pos, np.zeros((3, len(lonsl)))], axis=0),
            in_frame="ITRS",
            out_frame="TEME",
        )

        ax.plot(
            pos[0, :],
            pos[1, :],
            pos[2, :],
            alpha=alpha,
            linestyle="-",
            linewidth=linewidth,
            marker="",
            color=color,
        )

    for lon in lons:
        pos = frames.geodetic_to_ITRS(latsl, np.ones_like(latsl) * lon, np.zeros_like(latsl))
        pos = frames.convert(
            time,
            np.concatenate([pos, np.zeros((3, len(latsl)))], axis=0),
            in_frame="ITRS",
            out_frame="TEME",
        )
        ax.plot(
            pos[0, :],
            pos[1, :],
            pos[2, :],
            alpha=alpha,
            linestyle="-",
            linewidth=linewidth,
            marker="",
            color=color,
        )
    if hide_ax:
        # Hide grid lines
        ax.grid(False)

        # Hide axes ticks
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])
        ax.set_axis_off()

    return ax



#!/usr/bin/env python

"""Defines the required properties of a radar system and its components, including predefined instances."""

from .scans import Scan
from .tx_rx import Station, TX, RX, StationId

from .radars import get_radar, list_radars



#!/usr/bin/env python

"""Defines all the available propagators and the base-class that needs to be sub-classed to implement a custom propagator.
Such a subclass can then be used with every other functionality of SORTS.

"""

import importlib.util
from .base import Propagator

__all__ = [
    "Propagator",
]

from .kepler import Kepler, KeplerSettings

__all__.append("Kepler")
__all__.append("KeplerSettings")

if importlib.util.find_spec("sgp4") is not None:
    from .pysgp4 import Sgp4, Sgp4Settings
    __all__.append("Sgp4")

Rebound = None

# if importlib.util.find_spec("orekit") is not None:
#     from .orekit import Orekit
#
#     __all__.append("Orekit")
#
#
# if importlib.util.find_spec("rebound") is not None:
#     from .rebound import Rebound
#
#     __all__.append("Rebound")
#
# if importlib.util.find_spec("poliastro") is not None:
#     from .poliastro import TwoBody
#
#     __all__.append("TwoBody")



from ..radar import Radar

RADARS = dict()


def list_radars():
    """Returns a dict listing all avalible Radars and their variants"""
    return {key: list(val.keys()) for key, val in RADARS.items()}


def register_radar(name, variant, generator, override_ok=False):
    """Registers a new radar"""
    if name not in RADARS:
        RADARS[name] = {}

    if variant in RADARS[name] and not override_ok:
        raise ValueError(f"{name} with {variant=} already registered")

    RADARS[name][variant] = generator


def radar_generator(name, variant, override_ok=False):
    """Decorator to automatically register the radar beam generator."""

    def registrator_wrapper(generator):
        register_radar(name, variant, generator, override_ok=override_ok)
        return generator

    return registrator_wrapper


def get_radar(name, variant, *args, **kwargs) -> Radar:
    """Get a predefined radar instance from the available library of radars.


    Parameters
    ----------
    name : str
        Name of the radar system
    variant : any
        Variant identifier of the radar system
    *args : required
        Additional required positional arguments by the radar instance.
    **kwargs : optional
        Additional keyword arguments supplied to the radar instance.

    Returns
    -------
    sorts.radar.Radar
        A radar instance.

    """
    if name not in RADARS:
        raise ValueError(
            f'"{name}" radar not found. See available Radars:\n'
            + ", ".join([str(x) for x in RADARS])
        )

    if variant not in RADARS[name]:
        raise ValueError(
            f'"{variant}" variant not found. See available variants for {name}:\n'
            + ", ".join([str(x) for x in RADARS[name]])
        )

    radar_generator = RADARS[name][variant]
    radar = radar_generator(*args, **kwargs)
    return radar



#!/usr/bin/env python

"""Defines function for reading the MASTER catalog files.

"""

import numpy as np

from .population import Population


def master_catalog(
    input_file,
    mjd0=54952.0,
    sort=True,
    propagator=None,
    propagator_options={},
    propagator_args={},
):
    """Return the master catalog specified in the input file as a population instance. The catalog only contains the master sampling objects and not an actual realization of the population using the factor.

    **Note:** MASTER catalog files are usually (# TODO: confirm) in TEME frame

    The format of the input master files is:

        0. ID
        1. Factor
        2. Mass [kg]
        3. Diameter [m]
        4. m/A [kg/m2]
        5. a [km]
        6. e
        7. i [deg]
        8. RAAN [deg]
        9. AoP [deg]
        10. M [deg]


    :param str input_file: Path to the input MASTER file.
    :param bool sort: If :code:`True` sort according to diameters in descending order.
    :param float mjd0: The epoch of the catalog file in Modified Julian Days.
    :param PropagatorBase propagator: Propagator class pointer used for :class:`space_object.SpaceObject`.
    :param dict propagator_options: Propagator initialization keyword arguments.


    :return: Master catalog
    :rtype: population.Population
    """
    master_raw = np.genfromtxt(input_file)
    i = [0, 5, 6, 7, 8, 9, 10]
    var = ["oid", "a", "e", "i", "raan", "aop", "mu0"]

    master = Population(
        fields=[
            "oid",
            "a",
            "e",
            "i",
            "aop",
            "raan",
            "mu0",
            "mjd0",
            "A",
            "m",
            "d",
            "C_D",
            "C_R",
            "Factor",
            "MASTER-ID",
        ],
        dtypes=["int"] + ["float64"] * 13 + ["int"],
        space_object_fields=["A", "m", "d", "C_D", "C_R"],
        state_fields=["a", "e", "i", "aop", "raan", "mu0"],
        epoch_field={"field": "mjd0", "format": "mjd", "scale": "utc"},
        propagator=propagator,
        propagator_options=propagator_options,
        propagator_args=propagator_args,
    )

    master.allocate(master_raw.shape[0])

    for ind, key in zip(i, var):
        master.data[key] = master_raw[:, ind]

    master.data["a"] *= 1e3  # km to m
    master.data["mjd0"] = mjd0
    master.data["A"] = np.divide(master_raw[:, 2], master_raw[:, 4])
    master.data["m"] = master_raw[:, 2]
    master.data["d"] = master_raw[:, 3]
    master.data["C_D"] = 2.3
    master.data["C_R"] = 1.0
    master.data["Factor"] = master_raw[:, 1]
    master.data["MASTER-ID"] = master_raw[:, 0]

    diams = master_raw[:, 3]

    if sort:
        idxs = np.argsort(diams)[::-1]
    else:
        idxs = np.arange(len(diams))

    master.data = master.data[idxs]

    return master


def master_catalog_factor(
    master_base,
    copy=True,
    treshhold=0.01,
    seed=None,
):
    """Returns a random realization of the master population specified by the input file/population. In other words, each sampling object in the catalog is sampled a "factor" number of times with random mean anomalies to create the population.

    :param population.Population master_base: A master catalog consisting only of sampling objects. This catalog will be modified and the pointer to it returned.
    :param float treshhold: Diameter limit in meters below witch sampling objects are not included. Can be :code:`None` to skip filtering.
    :param int seed: Random number generator seed given to :code:`numpy.random.seed` to allow for consisted generation of a random realization of the population. If seed is :code:`None` a random seed from high-entropy data is used.
    :param bool copy: Modify the given `master_base` instance or return a modified copy.


    :return: Master population
    :rtype: population.Population
    """
    if seed is not None:
        st0 = np.random.get_state()
        np.random.seed(seed=seed)

    if copy:
        master = master_base.copy()
    else:
        master = master_base

    if treshhold is not None:
        master.filter("d", lambda d: d >= treshhold)

    full_objs = np.zeros(
        (int(np.sum(np.round(master["Factor"]))), master.shape[1]), dtype=np.float64
    )

    max_oid = np.max(master["oid"])
    max_factor = np.floor(np.log10(np.max(master["Factor"])))
    oid_extend = 10 ** (int(np.log10(max_oid)) + max_factor + 1.0)
    oid_mag = 10 ** (int(np.log10(max_oid)))

    i = 0
    for row in master.data:
        f_int = int(np.round(row[13]))
        if f_int >= 1:
            ip = i + f_int
            for coli, head in enumerate(master.fields):
                full_objs[i:ip, coli] = row[head]

            full_objs[i:ip, 0] = np.array(range(i, ip), dtype=np.float64)
            full_objs[i:ip, 6] = np.random.rand(f_int) * 360.0
            i = ip

    master.allocate(full_objs.shape[0])

    master[:, :] = full_objs

    if seed is not None:
        np.random.seed(None)
        np.random.set_state(st0)

    return master



#!/usr/bin/env python

"""TLE list loading to population

"""
import pathlib

import numpy as np
from astropy.time import Time
import pyorb

from ..propagator import Sgp4
from .population import Population


def tle_catalog(
    tles,
    sgp4_propagation=True,
    cartesian=True,
    kepler=False,
    propagator=None,
    propagator_options={},
    propagator_args={},
):
    """Reads a TLE-snapshot file and converts the TLE's to orbits in a TEME frame and creates a population file.
    A snapshot generally contains several TLE's for the same object thus will this population also contain duplicate objects.
    The BSTAR parameter is saved in field BSTAR :code:`BSTAR`.

    *Numerical propagator assumptions:*
    To propagate with a numerical propagator one needs to make assumptions.
       * Density is $5\\cdot 10^3 \\;\\frac{kg}{m^3}$.
       * Object is a sphere
       * Drag coefficient is 2.3.

    :param str/list tles: Path to the input TLE snapshot file. Or the TLE-set can be given directly as a list of two lines that can be unpacked in a loop, e.g. :code:`[(tle1_l1, tle1_l2), (tle2_l1, tle2_l2)]`.

    :return: TLE snapshot as a Population
    :rtype: sorts.Population
    """
    if isinstance(tles, str) or isinstance(tles, pathlib.Path):
        # first character in a line is line number (1 or 2), so just ignore everything else
        tle = {"1": [], "2": []}
        for line in open(tles):
            num = line[0]
            if num not in ["1", "2"]:
                continue
            tle[num].append(line.rstrip("\n"))

        if len(tle["1"]) != len(tle["2"]):
            raise Exception("Not even number of lines [not TLE compatible]")

        tles = list(zip(tle["1"], tle["2"]))

    tle_size = len(tles)

    prop = Sgp4(
        settings=dict(
            out_frame="TEME",
            tle_input=True,
        ),
    )

    cart_names = ["x", "y", "z", "vx", "vy", "vz"]
    kep_names = ["a", "e", "i", "aop", "raan", "mu0"]
    params = ["mjd0", "A", "m", "d", "C_D", "C_R", "BSTAR"]

    fields = ["oid"] + cart_names
    dtypes = ["int"] + ["float64"] * len(cart_names)
    if kepler:
        fields += kep_names
        dtypes += ["float64"] * len(kep_names)

    fields += params
    dtypes += ["float64"] * len(params)

    if sgp4_propagation:
        fields += ["line1", "line2"]
        dtypes += ["S70"] * 2

    if sgp4_propagation:
        pop = Population(
            fields=fields,
            dtypes=dtypes,
            space_object_fields=["m", "d"],
            state_fields=["line1", "line2"],
            epoch_field={"field": "mjd0", "format": "mjd", "scale": "utc"},
            propagator=SGP4,
            propagator_options=dict(
                settings=dict(
                    out_frame="TEME",
                    tle_input=True,
                ),
            ),
        )
    else:
        pop = Population(
            fields=fields,
            dtypes=dtypes,
            space_object_fields=["A", "m", "d", "C_D", "C_R", "BSTAR"],
            state_fields=["x", "y", "z", "vx", "vy", "vz"],
            epoch_field={"field": "mjd0", "format": "mjd", "scale": "utc"},
            propagator=propagator,
            propagator_options=propagator_options,
            propagator_args=propagator_args,
        )

    pop.allocate(tle_size)

    for line_id, lines in enumerate(tles):
        line1, line2 = lines

        params = SGP4.get_TLE_parameters(line1, line2)
        bstar = params["bstar"]
        epoch = Time(params["jdsatepoch"] + params["jdsatepochF"], format="jd", scale="utc").mjd
        oid = params["satnum"]

        if cartesian or kepler:
            state_TEME = prop.propagate([0.0], lines)

            if kepler:
                kep = pyorb.cart_to_kep(state_TEME, mu=pyorb.G * pyorb.M_earth, degrees=True)
                pop.data[line_id]["a"] = kep[0]
                pop.data[line_id]["e"] = kep[1]
                pop.data[line_id]["i"] = kep[2]
                pop.data[line_id]["aop"] = kep[3]
                pop.data[line_id]["raan"] = kep[4]
                pop.data[line_id]["mu0"] = pyorb.true_to_mean(kep[5], kep[1], degrees=True)

            pop.data[line_id]["x"] = state_TEME[0]
            pop.data[line_id]["y"] = state_TEME[1]
            pop.data[line_id]["z"] = state_TEME[2]
            pop.data[line_id]["vx"] = state_TEME[4]
            pop.data[line_id]["vy"] = state_TEME[3]
            pop.data[line_id]["vz"] = state_TEME[5]

        pop.data[line_id]["oid"] = oid
        pop.data[line_id]["mjd0"] = epoch

        if sgp4_propagation:
            pop.data[line_id]["line1"] = line1
            pop.data[line_id]["line2"] = line2

        bstar = bstar / (prop.grav_model.radiusearthkm * 1000.0)

        pop.data[line_id]["BSTAR"] = bstar

        B = bstar * 2.0 / prop.rho0
        if B < 1e-9:
            rho = 500.0
            C_D = 0.0
            r = 0.1
            A = np.pi * r**2
            m = rho * 4.0 / 3.0 * np.pi * r**3
        else:
            C_D = 2.3
            rho = 5.0
            r = (3.0 * C_D) / (B * rho)
            A = np.pi * r**2
            m = rho * 4.0 / 3.0 * np.pi * r**3

        pop.data[line_id]["A"] = A
        pop.data[line_id]["m"] = m
        pop.data[line_id]["d"] = r * 2.0
        pop.data[line_id]["C_D"] = C_D
        pop.data[line_id]["C_R"] = 1.0

    return pop



import argparse
import logging

logger = logging.getLogger(__name__)

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
except ImportError:

    class COMM_WORLD:
        rank = 0
        size = 1

    comm = COMM_WORLD()

COMMANDS = dict()


def build_parser():
    parser = argparse.ArgumentParser(
        description="Space Object Radar Tracking Simulator command line tool"
    )

    parser.add_argument(
        "-v", "--verbose", help="Increase output verbosity", action="count", default=0
    )

    subparsers = parser.add_subparsers(help="Avalible command line interfaces", dest="command")
    subparsers.required = True

    for name, dat in COMMANDS.items():
        parser_builder, command_help = dat["parser"]
        cmd_parser = subparsers.add_parser(name, help=command_help)
        parser_builder(cmd_parser)

    return parser


def add_command(name, function, parser_build, command_help=""):
    global COMMANDS
    COMMANDS[name] = dict()
    COMMANDS[name]["function"] = function
    COMMANDS[name]["parser"] = (parser_build, command_help)


def main():
    parser = build_parser()
    args = parser.parse_args()

    function = COMMANDS[args.command]["function"]

    function(args)



from __future__ import annotations
from tqdm import tqdm
import logging, typing as t, traceback, time, abc
import sys
from mpi4py import MPI


logger = logging.getLogger(__name__)


class WorkerError(RuntimeError):
    pass


class _MpiK:
    """Internal helper for accessing string constant consistently"""

    exit_ok: t.Final = "exit_ok"
    terminate: t.Final = "terminate"


# TODO:
# The architecture of this class could use a nice diagram, its really general purpose which is super
# nice! But to avoid tracing out the control flow manually, a diagram could shortcut someone looking
# at this for the first time

# NOTE: Resorted to using a loose `Mapping` type instead of setting a generic param of `MpiQueuedExecution`
#       because it seems python does not infer generic param based on method signatures of subclasses.
#       Which mean if the generic param are not provided when subclassing, they are considered as Any/Unknown.
#
#       Althought using a loosely typed dict will not provide a strict type correctness,
#       it at least provide the following values:
#       - a loosely typed dict is better than Unknown/Any
#       - user can cast it to TypedDict for enhanced type safety will minimal friction
#       - code are easier to read without generics, especially when the behaviour of subclassing with generics is not so clear
#
#       An alternative solution is to provide default values for generic param.
#       Such feature is recently added in python 3.13, which is released on October 7, 2024.
#       But at the moment of writing, we are on October 29, 2025 only
#       and 3.13 is too new to be adopted as the baseline python version for a lib.
WorkerJobParam = t.Mapping
"""Alias of `Mapping`. Cast to `TypedDict` for enhanced type safety."""


class MpiQueuedExecution(abc.ABC):
    """
    NOTE: In typical usage, an instance of this class will be created on each MPI process, which means:
        - Each instance will init it's attributes/properties independently at different time
        - Attributes/Properties of the same name can end up having different value
          (e.g. a timestamp attribute will have different value on each instance.)

    """

    master_proc_rank = 0

    def __init__(self, is_run_with_mpi=True, progress=False):
        self.is_run_with_mpi = is_run_with_mpi

        self.comm = MPI.COMM_WORLD
        self.rank = self.comm.Get_rank()
        self.num_workers = self.comm.Get_size() - 1
        self.progress = progress

    # TODO: maybe this can take kwargs that are passed from run?
    @abc.abstractmethod
    def master_process(self) -> None:
        """
        The code that only ran on the master rank process.
        - Must invoke the method `mpi_master_proc_loop` at least once to start dispatching job to workers.
        """

    @abc.abstractmethod
    def worker_process(self, job_param: WorkerJobParam) -> None:
        """The code that only ran on the worker rank processes."""
        ...

    def _mpi_master_proc_loop_with_mpi(self, work_job_params: t.Sequence[WorkerJobParam]) -> None:
        rank_size = self.comm.Get_size()

        logger.debug(f"running in mpi with rank: {rank_size}")
        logger.info(f"master: {self.master_proc_rank} | simulation preparation start")

        ##
        # parallization section
        ##
        logger.info(
            f"master: {self.master_proc_rank} | parallel processing of SimulationUnit start"
        )

        is_worker_idle_list = [True for _ in range(self.num_workers)]
        next_work_job_param_idx = 0
        processed_work_job_cnt = 0
        if self.progress:
            pbar = tqdm("MPI worker progress", total=len(work_job_params), file=sys.stdout)

        while processed_work_job_cnt < len(work_job_params):
            # send next work_job_param if there is idle worker
            if any(is_worker_idle_list) and next_work_job_param_idx < len(work_job_params):
                idle_worker_idx = is_worker_idle_list.index(True)
                idle_worker_rank = idle_worker_idx + 1

                job_param = work_job_params[next_work_job_param_idx]
                self.comm.send(job_param, dest=idle_worker_rank)

                logger.debug(
                    f"master: {self.master_proc_rank} | sent `SimulationUnit`"
                    + f" ({next_work_job_param_idx+1}/{len(work_job_params)}) to worker {idle_worker_rank}"
                )

                is_worker_idle_list[idle_worker_idx] = False
                next_work_job_param_idx += 1

            # otherwise, wait for result
            else:
                logger.debug(f"master: {self.master_proc_rank} | awaiting results ...")

                status = MPI.Status()
                self.comm.recv(status=status)
                worker_rank = status.Get_source()

                processed_work_job_cnt += 1
                if self.progress:
                    pbar.update(1)

                is_worker_idle_list[worker_rank - 1] = True

        if self.progress:
            pbar.close()
        logger.info(f"master: {self.master_proc_rank} | master proc loop done, returning...")

    def _mpi_master_proc_loop_without_mpi(
        self, work_job_params: t.Sequence[WorkerJobParam]
    ) -> None:
        """This will be ran instead of `mpi_master_proc_loop` when `is_run_with_mpi` is `False`"""

        if self.progress:
            pbar = tqdm("Worker progress", total=len(work_job_params), file=sys.stdout)
        for work_job_param in work_job_params:
            self.worker_process(work_job_param)
            if self.progress:
                pbar.update(1)
        if self.progress:
            pbar.close()
        logger.info("master proc loop done, returning...")

    def mpi_master_proc_loop(self, work_job_params: t.Sequence[WorkerJobParam]) -> None:
        if self.is_run_with_mpi:
            return self._mpi_master_proc_loop_with_mpi(work_job_params)
        else:
            return self._mpi_master_proc_loop_without_mpi(work_job_params)

    def mpi_worker_proc_loop(self) -> None:
        worker_proc_rank = self.rank

        while True:
            logger.info(f"worker: {worker_proc_rank} | waiting for msg...")
            msg = self.comm.recv(source=self.master_proc_rank)

            if msg == _MpiK.terminate:
                # exit if `_MpiK.terminate` is received
                logger.info(f"worker: {worker_proc_rank} | exiting...")
                self.comm.send(_MpiK.exit_ok, dest=self.master_proc_rank)  # reply an ack to master

                # break here to allow for further execution of workers later
                break

            elif isinstance(msg, t.Mapping):
                job_param = t.cast(WorkerJobParam, msg)
                try:
                    self.worker_process(job_param)
                except BaseException as exc:
                    raise WorkerError(
                        "Error during job:\n "
                        + "\n".join([f"{key}: {val}" for key, val in job_param.items()])
                    ) from exc

            else:
                # throw for unexpected msg
                raise RuntimeError(f"worker: {worker_proc_rank} | received unexcepted msg: {msg}")

    def _run_with_mpi(self):
        try:
            # master
            if self.rank == self.master_proc_rank:
                calc_start_time = time.perf_counter()

                self.master_process()

                # TODO: maybe we can use `comm.bcast` here?
                #   but worker also need to call `comm.bcast` for listening,
                #   not sure if mpi allows listening to both `bcast` and `recv`
                for r in range(1, self.num_workers + 1):
                    logger.debug(f"master: {self.master_proc_rank} | terminating worker: {r} ...")
                    self.comm.send(_MpiK.terminate, dest=r)
                    self.comm.recv(source=r)  # wait for an ack

                calc_time = time.perf_counter() - calc_start_time
                logger.info(f"master_process took {calc_time} sec")

            # workers
            else:
                self.mpi_worker_proc_loop()

        except Exception as err:
            comm = MPI.COMM_WORLD
            r = comm.Get_rank()

            logger.error(
                f"terminating mpi proc due to exception occured in rank: {r}, error:\n"
                + "\n".join(traceback.format_exception(err))
            )
            comm.Abort(1)

    def _run_without_mpi(self):
        """Run the execution without MPI, mostly useful for debugging."""

        calc_start_time = time.perf_counter()

        self.master_process()

        calc_time = time.perf_counter() - calc_start_time
        logger.info(f"master_process took {calc_time} sec")

    def run(self):
        if self.is_run_with_mpi:
            self._run_with_mpi()
        else:
            self._run_without_mpi()



#!/usr/bin/env python

"""Basis for developing a scheduler for tracking multiple space objects."""
import logging
from abc import abstractmethod

import numpy as np
import pyorb

from ..controller_v1 import Tracker
from .scheduler import Scheduler
from ..passes import equidistant_sampling

logger = logging.getLogger(__name__)


class Tracking(Scheduler):
    """#TODO: Docstring"""

    def __init__(
        self,
        radar,
        space_objects,
        end_time,
        epoch,
        start_time=0.0,
        controller=Tracker,
        controller_args=dict(return_copy=True),
        max_dpos=1e3,
        max_samp=30.0,
        use_pass_states=True,
        calculate_max_snr=False,
        collect_passes=False,
    ):
        super().__init__(radar)
        self.epoch = epoch
        self.controller = controller
        self.controller_args = controller_args
        self.space_objects = space_objects
        self.start_time = start_time
        self.end_time = end_time
        self.collect_passes = collect_passes

        self.max_dpos = max_dpos
        self.max_samp = max_samp
        self.use_pass_states = use_pass_states
        self.calculate_max_snr = calculate_max_snr

        self.reset_space_objects(space_objects)

    def reset_space_objects(self, space_objects):
        self.space_objects = space_objects
        self.states = [None] * len(space_objects)
        self.states_t = [None] * len(space_objects)
        self.passes = [None] * len(space_objects)

        if self.use_pass_states:
            self.measurements = [None] * len(space_objects)
        else:
            self.measurements = None

    def add_space_object(self, space_object):
        self.space_objects.append(space_object)
        self.states.append(None)
        self.states_t.append(None)
        self.passes.append(None)

        if self.use_pass_states:
            self.measurements.append(None)

    def update(self):
        self.reset_space_objects(self.space_objects)
        for ind in range(len(self.space_objects)):
            self.get_passes(ind)

    def get_passes(self, ind):
        """#TODO: Docstring"""
        dt = (self.space_objects[ind].epoch - self.epoch).to_value("sec")

        if isinstance(self.space_objects[ind].state, pyorb.Orbit):
            t = equidistant_sampling(
                orbit=self.space_objects[ind].state,
                start_t=self.start_time,
                end_t=self.end_time,
                max_dpos=self.max_dpos,
            )
        else:
            t = np.arange(self.start_time, self.end_time, self.max_samp)

        logger.info(f"Tracking:get_passes(ind={ind}):propagating {len(t)} steps")

        states = self.space_objects[ind].get_state(t - dt)
        if self.use_pass_states:
            self.states[ind] = states
            self.states_t[ind] = t

        logger.info(f"Tracking:get_passes(ind={ind}):propagating complete")

        self.passes[ind] = self.radar.find_passes(t, states, cache_data=True)

        # we may need SNR to plan observations
        for txi in range(len(self.radar.tx)):
            for rxi in range(len(self.radar.rx)):
                logger.info(
                    f"Tracking:get_passes(ind={ind}):tx{txi}-rx{rxi} {len(self.passes[ind][txi][rxi])} passes"
                )
                if not self.calculate_max_snr:
                    continue
                for ps in self.passes[ind][txi][rxi]:
                    ps.calculate_max_snr(
                        rx=self.radar.rx[rxi],
                        tx=self.radar.tx[txi],
                        diameter=self.space_objects[ind].d,
                    )

        return self.passes[ind], states, t

    @abstractmethod
    def set_measurements(self, *args, **kwargs):
        """#TODO: Docstring"""
        pass

    def get_controllers(self):
        """#TODO: Docstring"""
        logger.debug(f"Tracking:get_controllers")

        ctrls = []
        for ind in range(len(self.space_objects)):
            if self.collect_passes:
                if self.use_pass_states:
                    minds = self.measurements[ind]
                    states = self.states[ind][:3, minds]
                    t = self.states_t[ind][minds]
                else:
                    states = self.states[ind][:3, :]
                    t = self.states_t[ind][:]

                ctrl = self.controller(
                    radar=self.radar, t=t, t0=0.0, ecefs=states[:3, :], **self.controller_args
                )
                ctrl.meta["target"] = f"Object {self.space_objects[ind].object_id}"
                ctrls.append(ctrl)

            else:
                for pass_ind in range(len(self.states_t[ind])):
                    if self.use_pass_states:
                        minds = self.measurements[ind][pass_ind]
                        states = self.states[ind][pass_ind][:3, minds]
                        t = self.states_t[ind][pass_ind][minds]
                    else:
                        states = self.states[ind][pass_ind][:3, :]
                        t = self.states_t[ind][pass_ind][:]

                    ctrl = self.controller(
                        radar=self.radar, t=t, t0=0.0, ecefs=states[:3, :], **self.controller_args
                    )
                    ctrl.meta["target"] = f"Object {self.space_objects[ind].object_id}"
                    ctrl.meta["pass"] = pass_ind
                    ctrls.append(ctrl)

        return ctrls


class PriorityTracking(Tracking):
    """#TODO: Docstring"""

    def __init__(self, radar, space_objects, end_time, epoch, timeslice, allocation, **kwargs):
        self.priority = kwargs.pop("priority", None)
        super().__init__(radar, space_objects, end_time, epoch, **kwargs)

        self.timeslice = timeslice
        self.allocation = allocation

        if self.priority is None:
            self.priority = np.ones((len(self.space_objects),), dtype=np.float64)

    def add_space_object(self, space_object, priority):
        super().add_space_object(space_object)
        self.priority.append(priority)

    def update(self, space_objects=None, priority=None):
        if space_objects is not None:
            self.space_objects = space_objects
        if priority is not None:
            self.priority = priority
        assert len(self.priority) == len(self.space_objects), "Need a priority for each object"

        super().update()

    def set_measurements(self):
        """#TODO: Docstring"""
        if not self.collect_passes:
            raise NotImplementedError()

        total_measurements = int(self.allocation / self.timeslice)
        if not isinstance(self.priority, np.ndarray):
            pri = np.array(self.priority)
            pri = pri / pri.sum()
        else:
            pri = self.priority / self.priority.sum()

        # per object
        measurements = np.floor(pri * total_measurements).astype(np.int64)

        for ind in range(len(self.space_objects)):
            if self.use_pass_states:
                all_inds = []
                for txi in range(len(self.passes[ind])):
                    for rxi in range(len(self.passes[ind][txi])):
                        all_inds += [ps.inds for ps in self.passes[ind][txi][rxi]]
                if len(all_inds) == 0:
                    self.measurements[ind] = np.empty((0,), dtype=np.int64)
                    continue
                all_inds = np.concatenate(all_inds, axis=0)
                all_inds = np.unique(all_inds)

                # uniform distribution
                dt = int(len(all_inds) / measurements[ind])
                if dt == 0:
                    dt = 1
                all_inds = all_inds[::dt]
                self.measurements[ind] = all_inds
            else:
                raise NotImplementedError()



#!/usr/bin/env python

"""#TODO

"""

import logging
import numpy as np

from .radar_controller import RadarController

logger = logging.getLogger(__name__)

class Tracker(RadarController):
    """Takes in ECEF points and a time vector and creates a tracking control."""

    META_FIELDS = RadarController.META_FIELDS + [
        "dwell",
        "target",
    ]

    def __init__(
        self,
        radar,
        t,
        ecefs,
        t0=0.0,
        dwell=0.1,
        return_copy=True,
        meta=None,
        **kwargs
    ):
        super().__init__(radar, t=t, t0=t0, meta=meta, **kwargs)
        self.ecefs = ecefs
        self.dwell = dwell
        self.return_copy = return_copy

        logger.info(f"Tracker:init")

    @property
    def dwell(self):
        return self.t_slice

    @dwell.setter
    def dwell(self, val):
        self.t_slice = val

    def default_meta(self):
        dic = super().default_meta()
        dic["dwell"] = self.dwell
        return dic

    def point_radar(self, radar, ind):
        RadarController.point_tx_ecef(radar, self.ecefs[:3, ind])
        RadarController.point_rx_ecef(radar, self.ecefs[:3, ind])

    def generator(self, t):
        logger.debug(f"Tracker:generator: len(t) = {len(t)}")

        for ti in range(len(t)):
            if self.return_copy:
                radar = self.radar.copy()
            else:
                radar = self.radar

            dt = t[ti] - self.t
            check = np.logical_and(dt >= 0, dt < self.dwell)
            ind = np.argmax(check)
            meta = self.default_meta()

            RadarController.coh_integration(radar, self.dwell)

            self.toggle_stations(t[ti], radar)
            self.point_radar(radar, ind)

            yield radar, meta

        logger.debug(f"Tracker:generator:completed")



#!/usr/bin/env python

"""Scheduler base class.

"""
# Python standard import
import logging
from abc import ABC, abstractmethod


# Third party import
import numpy as np

# Local import

logger = logging.getLogger(__name__)

class Scheduler(ABC):
    """A Scheduler for executing time-slices of different radar controllers.

    #TODO: Docstring
    """

    def __init__(self, radar):
        self.radar = radar

    @abstractmethod
    def update(self, *args, **kwargs):
        """Update the scheduler information."""
        pass

    @abstractmethod
    def get_controllers(self):
        """This should init a list of controllers and set the `t` (global time) and `t0` (global time reference for controller) variables on them for their individual time samplings."""
        pass

    def generate_schedule(self, t, generator):
        """Takes times and a corresponding generator that returns radar instances to generate a radar schedule."""
        raise NotImplementedError()

    def calculate_observation(self, txrx_pass, t, generator, **kwargs):
        """Takes a pass over a tx-rx pair and the corresponding evaluated times and generator that returns radar instances to generate a set of observed data."""
        raise NotImplementedError()

    @staticmethod
    def chain_generators(generators):
        for generator in generators:
            yield from generator

    def schedule(self, start=None, stop=None, **kwargs):
        """Get the schedule generated by the self.generate_schedule method based on the times and generator."""
        ctrls = self.get_controllers()

        if start is not None and stop is not None:
            check_t = lambda c: np.logical_and(c.t >= start, c.t <= stop)
        elif start is not None:
            check_t = lambda c: c.t >= start
        elif stop is not None:
            check_t = lambda c: c.t <= stop
        else:
            check_t = lambda c: np.full(c.t.shape, True, dtype=bool)

        ctrls = [c for c in ctrls if np.any(check_t(c))]

        if len(ctrls) == 0:
            return self.generate_schedule(np.empty((0,)), [], **kwargs)

        times = np.concatenate([c.t[check_t(c)] for c in ctrls], axis=0)
        sched = Scheduler.chain_generators([c(c.t[check_t(c)] - c.t0) for c in ctrls])
        return self.generate_schedule(times, sched, **kwargs)

    def __call__(self, start, stop):
        """#TODO: Docstring"""
        ctrls = self.get_controllers()

        check_t = lambda c: np.logical_and(c.t >= start, c.t <= stop)

        ctrls = [c for c in ctrls if np.any(check_t(c))]
        if len(ctrls) == 0:
            logger.debug(
                f"Scheduler:__call__:No radar events found between {start:.1f} and {stop:.1f}"
            )
            return None, None
        else:
            t = np.concatenate([c.t[check_t(c)] for c in ctrls], axis=0)
            logger.debug(
                f"Scheduler:__call__:{len(t)} events found between {start:.1f} and {stop:.1f}"
            )

            return t, Scheduler.chain_generators([c(c.t[check_t(c)] - c.t0) for c in ctrls])

    def observe_passes(self, passes, **kwargs):
        """#TODO: Docstring"""

        linear_list = kwargs.pop("linear_list", False)

        data = []

        if linear_list:
            for ps in passes:
                t, generator = self(ps.start(), ps.end())
                if generator is not None:
                    pass_data = self.calculate_observation(ps, t, generator, **kwargs)
                else:
                    pass_data = None
                data.append(pass_data)
        else:
            for txi in range(len(passes)):
                data.append([])
                for rxi in range(len(passes[txi])):
                    data[-1].append([])
                    for ps in passes[txi][rxi]:
                        t, generator = self(ps.start(), ps.end())
                        if generator is not None:
                            pass_data = self.calculate_observation(ps, t, generator, **kwargs)
                        else:
                            pass_data = None
                        data[-1][-1].append(pass_data)
        return data



#!/usr/bin/env python

"""Coordinate frame transformations and related functions.
Main usage is the :code:`convert` function that wraps Astropy frame transformations.

"""

import logging
from collections import OrderedDict
import numpy as np
import astropy.coordinates as coord
import astropy.units as units
from astropy.coordinates import EarthLocation
from sorts.types import GeodeticCoordinates_DegM


try:
    from jplephem.spk import SPK
except ImportError:
    SPK = None

from spacecoords.spherical import sph_to_cart

logger = logging.getLogger(__name__)

"""List of astropy frames
"""
ASTROPY_FRAMES = {
    "TEME": "TEME",
    "ITRS": "ITRS",
    "ITRF": "ITRS",
    "ICRS": "ICRS",
    "ICRF": "ICRS",
    "GCRS": "GCRS",
    "GCRF": "GCRS",
    "HCRS": "HCRS",
    "HCRF": "HCRS",
    "HeliocentricMeanEcliptic".upper(): "HeliocentricMeanEcliptic",
    "GeocentricMeanEcliptic".upper(): "GeocentricMeanEcliptic",
    "HeliocentricTrueEcliptic".upper(): "HeliocentricTrueEcliptic",
    "GeocentricTrueEcliptic".upper(): "GeocentricTrueEcliptic",
    "BarycentricMeanEcliptic".upper(): "BarycentricMeanEcliptic",
    "BarycentricTrueEcliptic".upper(): "BarycentricTrueEcliptic",
    "SPICEJ2000": "ICRS",
}

ASTROPY_NOT_OBSTIME = [
    "ICRS",
    "BarycentricMeanEcliptic",
    "BarycentricTrueEcliptic",
]

"""Mapping from body name to integer id's used by the kernels.

Taken from `astropy.coordinates.solar_system`
"""
BODY_NAME_TO_KERNEL_SPEC = OrderedDict(
    [
        ("sun", [(0, 10)]),
        ("mercury", [(0, 1), (1, 199)]),
        ("venus", [(0, 2), (2, 299)]),
        ("earth-moon-barycenter", [(0, 3)]),
        ("earth", [(0, 3), (3, 399)]),
        ("moon", [(0, 3), (3, 301)]),
        ("mars", [(0, 4)]),
        ("jupiter", [(0, 5)]),
        ("saturn", [(0, 6)]),
        ("uranus", [(0, 7)]),
        ("neptune", [(0, 8)]),
        ("pluto", [(0, 9)]),
    ]
)


def not_geocentric(frame):
    """Check if the given frame name is one of the non-geocentric frames."""
    frame = frame.upper()
    return frame in ["ICRS", "ICRF", "HCRS", "HCRF"] or frame.startswith("Heliocentric".upper())


def is_geocentric(frame):
    """Check if the frame name is a supported geocentric frame"""
    return not not_geocentric(frame)


def arctime_to_degrees(minutes, seconds):
    return (minutes + seconds / 60.0) / 60.0


def get_solarsystem_body_states(bodies, epoch, kernel, units=None):
    """Open a kernel file and get the statates of the given bodies at epoch in ICRS.

    Note: All outputs from kernel computations are in the Barycentric (ICRS) "eternal" frame.
    """
    assert SPK is not None, "jplephem package needed to directly interact with kernels"
    states = {}

    kernel = SPK.open(kernel)

    epoch_ = epoch.tdb  # jplephem uses Barycentric Dynamical Time (TDB)
    jd1, jd2 = epoch_.jd1, epoch_.jd2

    for body in bodies:
        body_ = body.lower().strip()

        if body_ not in BODY_NAME_TO_KERNEL_SPEC:
            raise ValueError(f'Body name "{body}" not recognized')

        states[body] = np.zeros((6,), dtype=np.float64)

        # if there are multiple steps to go from states to
        # ICRS barycentric, iterate trough and combine
        for pair in BODY_NAME_TO_KERNEL_SPEC[body_]:
            spk = kernel[pair]
            if spk.data_type == 3:
                # Type 3 kernels contain both position and velocity.
                posvel = spk.compute(jd1, jd2).flatten()
            else:
                pos_, vel_ = spk.compute_and_differentiate(jd1, jd2)
                posvel = np.zeros((6,), dtype=np.float64)
                posvel[:3] = pos_
                posvel[3:] = vel_

            states[body] += posvel

        # units from kernels are usually in km and km/day
        if units is None:
            states[body] *= 1e3
            states[body][3:] /= 86400.0
        else:
            states[body] *= units[0]
            states[body][3:] /= units[1]

    return states


def convert(t, states, in_frame, out_frame, **kwargs):
    """Perform predefined coordinate transformations using Astropy.
    Always returns a copy of the array.

    :param numpy.ndarray/float t: Absolute time corresponding to the input states.
    :param numpy.ndarray states: Size `(6,n)` matrix of states in SI units where
        rows 1-3 are position and 4-6 are velocity.
    :param str in_frame: Name of the frame the input states are currently in.
    :param str out_frame: Name of the state to transform to.
    :rtype: numpy.ndarray
    :return: Size `(6,n)` matrix of states in SI units where rows
        1-3 are position and 4-6 are velocity.

    """

    logger.debug(f"frames:convert: in_frame={in_frame}, out_frame={out_frame}")

    in_frame = in_frame.upper()
    out_frame = out_frame.upper()

    if in_frame == out_frame:
        return states.copy()

    if in_frame in ASTROPY_FRAMES:
        in_frame_ = ASTROPY_FRAMES[in_frame]
        in_frame_cls = getattr(coord, in_frame_)
    else:
        err_str = (
            f"In frame '{in_frame}' not recognized, "
            "please check spelling or perform manual transformation"
        )
        raise ValueError(err_str)

    kw = {}
    kw.update(kwargs)
    if in_frame_ not in ASTROPY_NOT_OBSTIME:
        kw["obstime"] = t

    astropy_states = _convert_to_astropy(states, in_frame_cls, **kw)

    if out_frame in ASTROPY_FRAMES:
        out_frame_ = ASTROPY_FRAMES[out_frame]
        out_frame_cls = getattr(coord, out_frame_)
    else:
        err_str = (
            f"Out frame '{out_frame}' not recognized, "
            "please check spelling or perform manual transformation"
        )
        raise ValueError(err_str)

    kw = {}
    kw.update(kwargs)
    if out_frame_ not in ASTROPY_NOT_OBSTIME:
        kw["obstime"] = t

    out_states = astropy_states.transform_to(out_frame_cls(**kw))

    rets = states.copy()
    rets[:3, ...] = out_states.cartesian.xyz.to(units.m).value
    rets[3:, ...] = out_states.velocity.d_xyz.to(units.m / units.s).value

    logger.debug("frames:convert:completed")

    return rets


def geodetic_to_ITRS(lat, lon, alt, degrees=True, ellipsoid=None):
    """Use `astropy.coordinates.EarthLocation` to transform from geodetic to ITRS."""

    if degrees:
        lat, lon = np.radians(lat), np.radians(lon)

    cord = EarthLocation.from_geodetic(
        lon=lon * units.rad,
        lat=lat * units.rad,
        height=alt * units.m,
        ellipsoid=ellipsoid,
    )
    x, y, z = cord.to_geocentric()

    if x.size <= 1:
        pos = np.empty((3,), dtype=np.float64)
    else:
        pos = np.empty((3, x.size), dtype=np.float64)

    pos[0, ...] = x.to(units.m).value
    pos[1, ...] = y.to(units.m).value
    pos[2, ...] = z.to(units.m).value

    return pos


def ITRS_to_geodetic(x, y, z, degrees=True, ellipsoid=None) -> GeodeticCoordinates_DegM:
    """Use `astropy.coordinates.EarthLocation` to transform from geodetic to ITRS.

    :param float x: X-coordinate in ITRS
    :param float y: Y-coordinate in ITRS
    :param float z: Z-coordinate in ITRS
    :param bool radians: If :code:`True` then all values are given in radians instead of degrees.
    :param str/None ellipsoid: Name of the ellipsoid model used for geodetic
    coordinates, for default value see Astropy `EarthLocation`.
    :rtype: numpy.ndarray
    :return: (3,) array of (latitude, longitude, height) above ellipsoid
    """

    cord = EarthLocation.from_geocentric(
        x=x * units.m,
        y=y * units.m,
        z=z * units.m,
    )
    lon, lat, height = cord.to_geodetic(ellipsoid=ellipsoid)

    # TODO: confirm with daniel if the modificaiton is okay (made it work with ndarray instead of just scala)
    cord_len = len(cord)
    if cord_len > 1:
        llh = np.empty((3, cord_len), dtype=np.float64)
    else:
        llh = np.empty((3,), dtype=np.float64)

    if degrees:
        u_ = units.deg
    else:
        u_ = units.rad
    llh[0] = lat.to(u_).value
    llh[1] = lon.to(u_).value
    llh[2] = height.to(units.m).value

    return llh


def _convert_to_astropy(states, frame, **kw):
    state_p = coord.CartesianRepresentation(states[:3, ...] * units.m)
    state_v = coord.CartesianDifferential(states[3:, ...] * units.m / units.s)
    astropy_states = frame(state_p.with_differentials(state_v), **kw)
    return astropy_states


def enu_to_ecef(lat, lon, alt, enu, degrees=True):
    """ENU (east/north/up) using geocentric zenith to ECEF coordinate system
    conversion, not including translation.

    :param float lat: Latitude of the origin in geocentric spherical coordinates
    :param float lon: Longitude of the origin in geocentric spherical coordinates
    :param float alt: Altitude above ellipsoid, **Unused in this implementation**.
    :param numpy.ndarray enu: (3,n) input matrix of positions in the ENU-convention.
    :param bool radians: If :code:`True` then all values are given in radians instead of degrees.
    :rtype: numpy.ndarray
    :return: (3,n) array x,y and z coordinates in ECEF.
    """
    if degrees:
        lat, lon = np.radians(lat), np.radians(lon)

    mx = np.array(
        [
            [-np.sin(lon), -np.sin(lat) * np.cos(lon), np.cos(lat) * np.cos(lon)],
            [np.cos(lon), -np.sin(lat) * np.sin(lon), np.cos(lat) * np.sin(lon)],
            [0, np.cos(lat), np.sin(lat)],
        ]
    )

    ecef = np.dot(mx, enu)
    return ecef


def ned_to_ecef(lat, lon, alt, ned, degrees=True):
    """NED (north/east/down) using geocentric zenith to ECEF coordinate system
    conversion, not including translation.

    :param float lat: Latitude of the origin in geocentric spherical coordinates
    :param float lon: Longitude of the origin in geocentric spherical coordinates
    :param float alt: Altitude **Unused in this implementation**.
    :param numpy.ndarray ned: (3,n) input matrix of positions in the NED-convention.
    :param bool radians: If :code:`True` then all values are given in radians instead of degrees.
    :rtype: numpy.ndarray
    :return: (3,n) array x,y and z coordinates in ECEF.
    """
    enu = np.empty(ned.size, dtype=ned.dtype)
    enu[0, ...] = ned[1, ...]
    enu[1, ...] = ned[0, ...]
    enu[2, ...] = -ned[2, ...]
    return enu_to_ecef(lat, lon, alt, enu, degrees=degrees)


def ecef_to_enu(lat, lon, alt, ecef, degrees=True):
    """ECEF coordinate system to local ENU (east,north,up) using geocentric
    zenith, not including translation.

    :param float lat: Latitude of the origin in geocentric spherical coordinates
    :param float lon: Longitude of the origin in geocentric spherical coordinates
    :param float alt: Altitude **Unused in this implementation**.
    :param numpy.ndarray ecef: (3,n) array x,y and z coordinates in ECEF.
    :param bool radians: If :code:`True` then all values are given in radians instead of degrees.
    :rtype: numpy.ndarray
    :return: (3,n) array x,y and z in local coordinates in the
        ENU-convention using geocentric zenith.
    """
    if degrees:
        lat, lon = np.radians(lat), np.radians(lon)

    mx = np.array(
        [
            [-np.sin(lon), -np.sin(lat) * np.cos(lon), np.cos(lat) * np.cos(lon)],
            [np.cos(lon), -np.sin(lat) * np.sin(lon), np.cos(lat) * np.sin(lon)],
            [0, np.cos(lat), np.sin(lat)],
        ]
    )
    enu = np.dot(np.linalg.inv(mx), ecef)
    return enu


def azel_to_ecef(lat, lon, alt, az, el, degrees=True):
    """Radar pointing (az,el) using geocentric zenith to unit vector in
    ECEF, not including translation.

    TODO: Docstring
    """
    shape = (3,)

    if isinstance(az, np.ndarray):
        if len(az.shape) == 0:
            az = float(az)
        elif len(az) > 1:
            shape = (3, len(az))
            az = az.flatten()
        else:
            az = az[0]

    if isinstance(el, np.ndarray):
        if len(el.shape) == 0:
            el = float(el)
        elif len(el) > 1:
            shape = (3, len(el))
            el = el.flatten()
        else:
            el = el[0]

    sph = np.empty(shape, dtype=np.float64)
    sph[0, ...] = az
    sph[1, ...] = el
    sph[2, ...] = 1.0
    enu = sph_to_cart(sph, degrees=degrees)
    return enu_to_ecef(lat, lon, alt, enu, degrees=degrees)


def vec_to_vec(vec_in, vec_out):
    """Get the rotation matrix that rotates `vec_in` to `vec_out` along the
    plane containing both. Uses quaternion calculations.
    """

    N = len(vec_in)
    if N != len(vec_out):
        raise ValueError("Input and output vectors must be same dimensionality.")

    assert N == 3, "Only implemented for 3d vectors"

    a = vec_in / np.linalg.norm(vec_in)
    b = vec_out / np.linalg.norm(vec_out)

    adotb = np.dot(a, b)
    axb = np.cross(a, b)
    axb_norm = np.linalg.norm(axb)

    # rotation in the plane frame of `vec_in` and `vec_out`
    G = np.zeros((N, N), dtype=vec_in.dtype)
    G[0, 0] = adotb
    G[0, 1] = -axb_norm
    G[1, 0] = axb_norm
    G[1, 1] = adotb
    G[2, 2] = 1

    # inverse of change of basis from standard orthonormal to `vec_in` and `vec_out` plane
    F = np.zeros((N, N), dtype=vec_in.dtype)
    F[:, 0] = a
    F[:, 1] = (b - adotb * a) / np.linalg.norm(b - adotb * a)
    F[:, 2] = axb

    # go to frame, rotation in plane, leave frame
    R = F @ G @ np.linalg.inv(F)

    return R
